#!/usr/bin/env python
#
# This example demonstrates the creation of multiple actors and the
# manipulation of their properties and transformations. It is a
# derivative of Cone.py, see that example for more information.
#

import vtk
import time

# 
# Next we create an instance of vtkConeSource and set some of its
# properties. The instance of vtkConeSource "cone" is part of a visualization
# pipeline (it is a source process object); it produces data (output type is
# vtkPolyData) which other filters may process.
#
cone = vtk.vtkConeSource ()
cone.SetHeight( 3.0 )
cone.SetRadius( 1.0 )
cone.SetResolution( 10 )

# 
# In this example we terminate the pipeline with a mapper process object.
# (Intermediate filters such as vtkShrinkPolyData could be inserted in
# between the source and the mapper.)  We create an instance of
# vtkPolyDataMapper to map the polygonal data into graphics primitives. We
# connect the output of the cone souece to the input of this mapper.
#
coneMapper = vtk.vtkPolyDataMapper()
coneMapper.SetInputConnection(cone.GetOutputPort())

# 
# Create an actor to represent the first cone. The actor's properties are
# modified to give it different surface properties. By default, an actor
# is create with a property so the GetProperty() method can be used.
#
coneActor = vtk.vtkActor()
coneActor.SetMapper(coneMapper)
coneActor.GetProperty().SetColor(0.2, 0.63, 0.79)
coneActor.GetProperty().SetDiffuse(0.7)
coneActor.GetProperty().SetSpecular(0.4)
coneActor.GetProperty().SetSpecularPower(20)

#
# Create a property and directly manipulate it. Assign it to the
# second actor.
#
property = vtk.vtkProperty()
property.SetColor(1.0, 0.3882, 0.2784)
property.SetDiffuse(0.7)
property.SetSpecular(0.4)
property.SetSpecularPower(20)

#
# Create a second actor and a property. The property is directly
# manipulated and then assigned to the actor. In this way, a single
# property can be shared among many actors. Note also that we use the
# same mapper as the first actor did. This way we avoid duplicating
# geometry, which may save lots of memory if the geoemtry is large.
coneActor2 = vtk.vtkActor()
coneActor2.SetMapper(coneMapper)
# coneActor2.GetProperty().SetColor(0.2, 0.63, 0.79)
coneActor2.SetProperty(property)
coneActor2.SetPosition(0, 2, 0)

#
# Create the Renderer and assign actors to it. A renderer is like a
# viewport. It is part or all of a window on the screen and it is responsible
# for drawing the actors it has.  We also set the background color here.
#
ren1 = vtk.vtkRenderer()
ren1.AddActor(coneActor)
ren1.AddActor(coneActor2)
ren1.SetBackground(0.1, 0.2, 0.4)

#
# Finally we create the render window which will show up on the screen
# We put our renderer into the render window using AddRenderer. We also
# set the size to be 300 pixels by 300.
#
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren1)
renWin.SetSize(300, 300)

#
# Now we loop over 360 degreeees and render the cone each time.
#
# for i in range(0,360):
#     time.sleep(0.03)
    
#     renWin.Render()
#     ren1.GetActiveCamera().Azimuth( 1 )




###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This module defines the following classes:
    - QShellDialog
    - QShell

QShell is based on ideas and code from PyCute developed by Gerard Vermeulen.
Used with the author's permission.
More information on PyCute, visit:
http://gerard.vermeulen.free.fr/html/pycute-intro.html

"""
from __future__ import division

from PyQt4 import QtGui, QtCore
import sys

from vistrails.core.bundles import py_import
from vistrails.core.interpreter.default import get_default_interpreter
from vistrails.gui.vistrails_palette import QVistrailsPaletteInterface

################################################################################

_shell_dialog = None

def get_shell_dialog():
    global _shell_dialog

    if _shell_dialog is not None:
        return _shell_dialog

    try:
        deps = {'pip': 'ipython>=1.0',
                'linux-ubuntu': 'ipython-qtconsole',
                'linux-debian': 'ipython-qtconsole'}

        IPython = py_import('IPython.qt.console.rich_ipython_widget', deps,
                            True)
        RichIPythonWidget = \
                IPython.qt.console.rich_ipython_widget.RichIPythonWidget
        py_import('IPython.qt.inprocess', deps, True)
        QtInProcessKernelManager = \
                IPython.qt.inprocess.QtInProcessKernelManager
    except ImportError:
        return None

    km = QtInProcessKernelManager()
    km.start_kernel()
    kernel = km.kernel
    kernel.gui = 'qt4'

    kernel_client = km.client()
    kernel_client.start_channels()

    class IPythonDialog(RichIPythonWidget, QVistrailsPaletteInterface):
        """This class incorporates an  IPython shell into a dockable widget for use in the
        VisTrails environment"""
        def __init__(self, parent=None):
            RichIPythonWidget.__init__(self, parent)
            self.old_streams = None
            self.running_workflow = False
            self.kernel_manager = km
            self.kernel_client = kernel_client
            self.exit_requested.connect(self.stop)
            self.setWindowTitle("Console")
            self.vistrails_interpreter = get_default_interpreter()

        def visibility_changed(self, visible):
            QVistrailsPaletteInterface.visibility_changed(self, visible)
            if visible:
                self.show()
            else:
                self.hide()
        def stop(self):
            kernel_client.stop_channels()
            km.shutdown_kernel()

        def hide(self):
            """suspend() -> None
            Called when hiding the parent window in order to recover the previous
            state.

            """
            #recovering the state
            if self.old_streams is not None:
                sys.stdout, sys.stderr, sys.stdin = self.old_streams
                self.old_streams = None
            RichIPythonWidget.hide(self)

        def show(self):
            """show() -> None
            Store previous state and starts capturing all interactive input and
            output.

            """
            # capture all interactive input/output
            if self.old_streams is None:
                self.old_streams = sys.stdout, sys.stderr, sys.stdin
                sys.stdout   = self
                sys.stderr   = self
                sys.stdin    = self
            RichIPythonWidget.show(self)

        def showEvent(self, e):
            """showEvent(e) -> None
            Event handler called when the dialog acquires focus

            """
            self.show()

        def flush(self):
            """flush() -> None.
            Simulate stdin, stdout, and stderr.

            """
            pass

        def isatty(self):
            """isatty() -> int
            Simulate stdin, stdout, and stderr.

            """
            return 1

        def readline(self):
            """readline() -> str

            Simulate stdin, stdout, and stderr.

            """
            return ""

        def write(self, text):
            """write(text: str) -> None
            Simulate stdin, stdout, and stderr.

            """
            self.input_buffer = ''
            if not self.running_workflow:
                self.running_workflow = True
                # make text blue
                self._append_plain_text("\n\x1b[34m<STANDARD OUTPUT>\x1b[0m\n", True)
            self._append_plain_text(text, True)
            self._prompt_pos = self._get_end_cursor().position()
            self._control.ensureCursorVisible()
            self._control.moveCursor(QtGui.QTextCursor.End)

        def eventFilter(self, obj, event):
            """ Reimplemented to ensure a console-like behavior in the underlying
                text widgets.
            """
            etype = event.type()
            if etype == QtCore.QEvent.KeyPress:
                self.running_workflow = False
            return RichIPythonWidget.eventFilter(self, obj, event)

    _shell_dialog = IPythonDialog
    return IPythonDialog



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""This file describes the analogy structure the spreadsheet holding and it
should be reimplemented to integrate between the spreadsheet and the analogy
"""

from __future__ import division

import vistrails.core.analogy
import vistrails.gui


class SpreadsheetAnalogyObject(object):
    """
    SpreadsheetAnalogyObject provides a API functions to integrate
    between the spreadsheet and the analogy. There will be only one
    instance of the analogy object

    """
    def __init__(self):
        """ SpreadsheetAnalogyObject() -> SpreadsheetAnalogyObject
        Initialize object properties

        """
        pass

    def isValid(self):
        """ isValid() -> bool
        Is the current analogy object is valid and can be applied to
        other pipeline

        """
        return True

    def createAnalogy(self, p1Info, p2Info):
        """ createAnalogy(p1Info: tuple, p2Info) -> bool
        p1Info, p2Info: (vistrailName, versionNumber, actions, pipeline)

        Setup an analogy object from p1 to p2 given their info in
        p1Info and p2Info. CAUTION: sometimes the actual 'pipeline' on
        the spreadsheet is different than the one created from the
        vistrailName:versionNumber. For example, the parameter
        exploration.  If that is the case, actions is a list of all
        actions has performed on the original pipeline to get to the
        pipeline.

        This function should return a boolean saying if the analogy
        has been successfully created or not.

        """
#         (p1_vistrail, p1_number, p1_actions, p1_pipeline) = p1Info
#         (p2_vistrail, p2_number, p2_actions, p2_pipeline) = p2Info
        self._p1Info = p1Info
        self._p2Info = p2Info
        return True

    def applyAnalogy(self, pInfo):
        """ createAnalogy(pInfo: tuple) -> pInfo
        pInfo: (vistrailName, versionNumber, actions, pipeline)

        Given the analogy, apply on the pipeline in pInfo and return
        the modified pipeline as a result. Along with the pipeline is
        the list of actions that has been applied. If no actions
        given (i.e. []), this can not be put back to the builder. If
        analogy is not applicable, this should return None

        """

        (p1_locator, p1_number, p1_actions, p1_pipeline, p1_controller) = self._p1Info
        (p2_locator, p2_number, p2_actions, p2_pipeline, p2_controller) = self._p2Info
        (p3_locator, p3_number, p3_actions, p3_pipeline, p3_controller) = pInfo

        if (p1_controller != p2_controller or
            p1_controller != p3_controller or
            p1_actions or
            p2_actions or
            p3_actions):
            return None

        perform = vistrails.core.analogy.perform_analogy_on_vistrail

        controller = p1_controller
        vt = controller.vistrail
        if controller.current_version != p3_number:
            controller.change_selected_version(p3_number)
        action = perform( vt, p1_number, p2_number, p3_number)

        controller.add_new_action(action)
        controller.perform_action(action)

        new_version = controller.current_version
        new_pipeline = vt.getPipeline(new_version)
        return (controller.locator, new_version, [], new_pipeline, controller)

    def __call__(self):
        """ __call__() -> SpreadsheetAnalogy
        Return self for calling method

        """
        return self

SpreadsheetAnalogy = SpreadsheetAnalogyObject()



# pragma: no testimport
try:
    import core.packagemanager
except ImportError:
    pass
else:
    raise AssertionError("Importing without prefix from outside the package "
                         "didn't raise ImportError")



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

"""generated automatically by auto_dao.py"""
from __future__ import division

from xml_dao import XMLDAO
from vistrails.db.versions.v0_6_0.domain import *

class DBPortSpecXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'portSpec':
            return None
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        name = self.convertFromStr(self.getAttribute(node, 'name'), 'str')
        type = self.convertFromStr(self.getAttribute(node, 'type'), 'str')
        spec = self.convertFromStr(self.getAttribute(node, 'spec'), 'str')
        
        obj = DBPortSpec(id=id,
                         name=name,
                         type=type,
                         spec=spec)
        obj.is_dirty = False
        return obj
    
    def toXML(self, portSpec, doc, node):
        if node is None:
            node = doc.createElement('portSpec')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(portSpec.db_id, 'long'))
        node.setAttribute('name',self.convertToStr(portSpec.db_name, 'str'))
        node.setAttribute('type',self.convertToStr(portSpec.db_type, 'str'))
        node.setAttribute('spec',self.convertToStr(portSpec.db_spec, 'str'))
        
        return node

class DBModuleXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'module':
            return None
        
        functions = []
        
        annotations = {}
        portSpecs = {}
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        cache = self.convertFromStr(self.getAttribute(node, 'cache'), 'int')
        name = self.convertFromStr(self.getAttribute(node, 'name'), 'str')
        package = self.convertFromStr(self.getAttribute(node, 'package'), 'str')
        version = self.convertFromStr(self.getAttribute(node, 'version'), 'str')
        
        location = None
        
        # read children
        for child in list(node.childNodes):
            if child.nodeName == 'location':
                location = self.getDao('location').fromXML(child)
            elif child.nodeName == 'function':
                function = self.getDao('function').fromXML(child)
                functions.append(function)
            elif child.nodeName == 'annotation':
                annotation = self.getDao('annotation').fromXML(child)
                annotations[annotation.db_id] = annotation
            elif child.nodeName == 'portSpec':
                portSpec = self.getDao('portSpec').fromXML(child)
                portSpecs[portSpec.db_id] = portSpec
            elif child.nodeType == child.TEXT_NODE and child.nodeValue.strip() == '':
                # node.removeChild(child)
                pass
            elif child.nodeType != child.TEXT_NODE:
                print '*** ERROR *** nodeName = %s' % child.nodeName
        
        obj = DBModule(id=id,
                       cache=cache,
                       name=name,
                       package=package,
                       version=version,
                       location=location,
                       functions=functions,
                       annotations=annotations,
                       portSpecs=portSpecs)
        obj.is_dirty = False
        return obj
    
    def toXML(self, module, doc, node):
        if node is None:
            node = doc.createElement('module')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(module.db_id, 'long'))
        node.setAttribute('cache',self.convertToStr(module.db_cache, 'int'))
        node.setAttribute('name',self.convertToStr(module.db_name, 'str'))
        node.setAttribute('package',self.convertToStr(module.db_package, 'str'))
        node.setAttribute('version',self.convertToStr(module.db_version, 'str'))
        
        # load DOM node map
        nodeMap = {}
        for childNode in node.childNodes:
            if childNode.nodeType == childNode.ELEMENT_NODE and self.hasAttribute(childNode, 'id'):
                nodeId = self.convertFromStr(self.getAttribute(childNode, 'id'), 'long')
                nodeMap[(childNode.nodeName, nodeId)] = childNode
        
        # set elements
        location = module.db_location
        if location is not None:
            if nodeMap.has_key(('location', location.db_id)):
                childNode = nodeMap[('location', location.db_id)]
                del nodeMap[('location', location.db_id)]
            else:
                childNode = doc.createElement('location')
                node.appendChild(childNode)
            self.getDao('location').toXML(location, doc, childNode)
        functions = module.db_functions
        for function in functions:
            if nodeMap.has_key(('function', function.db_id)):
                childNode = nodeMap[('function', function.db_id)]
                del nodeMap[('function', function.db_id)]
            else:
                childNode = doc.createElement('function')
                node.appendChild(childNode)
            self.getDao('function').toXML(function, doc, childNode)
        annotations = module.db_annotations
        for annotation in annotations.itervalues():
            if nodeMap.has_key(('annotation', annotation.db_id)):
                childNode = nodeMap[('annotation', annotation.db_id)]
                del nodeMap[('annotation', annotation.db_id)]
            else:
                childNode = doc.createElement('annotation')
                node.appendChild(childNode)
            self.getDao('annotation').toXML(annotation, doc, childNode)
        portSpecs = module.db_portSpecs
        for portSpec in portSpecs.itervalues():
            if nodeMap.has_key(('portSpec', portSpec.db_id)):
                childNode = nodeMap[('portSpec', portSpec.db_id)]
                del nodeMap[('portSpec', portSpec.db_id)]
            else:
                childNode = doc.createElement('portSpec')
                node.appendChild(childNode)
            self.getDao('portSpec').toXML(portSpec, doc, childNode)
        
        # delete nodes not around anymore
        for childNode in nodeMap.itervalues():
            childNode.parentNode.removeChild(childNode)
        return node

class DBTagXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'tag':
            return None
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        name = self.convertFromStr(self.getAttribute(node, 'name'), 'str')
        
        obj = DBTag(id=id,
                    name=name)
        obj.is_dirty = False
        return obj
    
    def toXML(self, tag, doc, node):
        if node is None:
            node = doc.createElement('tag')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(tag.db_id, 'long'))
        node.setAttribute('name',self.convertToStr(tag.db_name, 'str'))
        
        return node

class DBPortXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'port':
            return None
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        type = self.convertFromStr(self.getAttribute(node, 'type'), 'str')
        moduleId = self.convertFromStr(self.getAttribute(node, 'moduleId'), 'long')
        moduleName = self.convertFromStr(self.getAttribute(node, 'moduleName'), 'str')
        sig = self.convertFromStr(self.getAttribute(node, 'sig'), 'str')
        
        obj = DBPort(id=id,
                     type=type,
                     moduleId=moduleId,
                     moduleName=moduleName,
                     sig=sig)
        obj.is_dirty = False
        return obj
    
    def toXML(self, port, doc, node):
        if node is None:
            node = doc.createElement('port')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(port.db_id, 'long'))
        node.setAttribute('type',self.convertToStr(port.db_type, 'str'))
        node.setAttribute('moduleId',self.convertToStr(port.db_moduleId, 'long'))
        node.setAttribute('moduleName',self.convertToStr(port.db_moduleName, 'str'))
        node.setAttribute('sig',self.convertToStr(port.db_sig, 'str'))
        
        return node

class DBLogXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'log':
            return None
        
        workflow_execs = {}
        machines = {}
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        
        
        # read children
        for child in list(node.childNodes):
            if child.nodeName == 'workflowExec':
                workflow_exec = self.getDao('workflow_exec').fromXML(child)
                workflow_execs[workflow_exec.db_id] = workflow_exec
            elif child.nodeName == 'machine':
                machine = self.getDao('machine').fromXML(child)
                machines[machine.db_id] = machine
            elif child.nodeType == child.TEXT_NODE and child.nodeValue.strip() == '':
                # node.removeChild(child)
                pass
            elif child.nodeType != child.TEXT_NODE:
                print '*** ERROR *** nodeName = %s' % child.nodeName
        
        obj = DBLog(id=id,
                    workflow_execs=workflow_execs,
                    machines=machines)
        obj.is_dirty = False
        return obj
    
    def toXML(self, log, doc, node):
        if node is None:
            node = doc.createElement('log')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(log.db_id, 'long'))
        
        # load DOM node map
        nodeMap = {}
        for childNode in node.childNodes:
            if childNode.nodeType == childNode.ELEMENT_NODE and self.hasAttribute(childNode, 'id'):
                nodeId = self.convertFromStr(self.getAttribute(childNode, 'id'), 'long')
                nodeMap[(childNode.nodeName, nodeId)] = childNode
        
        # set elements
        workflow_execs = log.db_workflow_execs
        for workflow_exec in workflow_execs.itervalues():
            if nodeMap.has_key(('workflow_exec', workflow_exec.db_id)):
                childNode = nodeMap[('workflow_exec', workflow_exec.db_id)]
                del nodeMap[('workflow_exec', workflow_exec.db_id)]
            else:
                childNode = doc.createElement('workflow_exec')
                node.appendChild(childNode)
            self.getDao('workflow_exec').toXML(workflow_exec, doc, childNode)
        machines = log.db_machines
        for machine in machines.itervalues():
            if nodeMap.has_key(('machine', machine.db_id)):
                childNode = nodeMap[('machine', machine.db_id)]
                del nodeMap[('machine', machine.db_id)]
            else:
                childNode = doc.createElement('machine')
                node.appendChild(childNode)
            self.getDao('machine').toXML(machine, doc, childNode)
        
        # delete nodes not around anymore
        for childNode in nodeMap.itervalues():
            childNode.parentNode.removeChild(childNode)
        return node

class DBMachineXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'machine':
            return None
        
        module_execs = []
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        name = self.convertFromStr(self.getAttribute(node, 'name'), 'str')
        os = self.convertFromStr(self.getAttribute(node, 'os'), 'str')
        architecture = self.convertFromStr(self.getAttribute(node, 'architecture'), 'str')
        processor = self.convertFromStr(self.getAttribute(node, 'processor'), 'str')
        ram = self.convertFromStr(self.getAttribute(node, 'ram'), 'int')
        
        obj = DBMachine(id=id,
                        name=name,
                        os=os,
                        architecture=architecture,
                        processor=processor,
                        ram=ram,
                        module_execs=module_execs)
        obj.is_dirty = False
        return obj
    
    def toXML(self, machine, doc, node):
        if node is None:
            node = doc.createElement('machine')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(machine.db_id, 'long'))
        node.setAttribute('name',self.convertToStr(machine.db_name, 'str'))
        node.setAttribute('os',self.convertToStr(machine.db_os, 'str'))
        node.setAttribute('architecture',self.convertToStr(machine.db_architecture, 'str'))
        node.setAttribute('processor',self.convertToStr(machine.db_processor, 'str'))
        node.setAttribute('ram',self.convertToStr(machine.db_ram, 'int'))
        
        return node

class DBAddXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'add':
            return None
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        what = self.convertFromStr(self.getAttribute(node, 'what'), 'str')
        objectId = self.convertFromStr(self.getAttribute(node, 'objectId'), 'long')
        parentObjId = self.convertFromStr(self.getAttribute(node, 'parentObjId'), 'long')
        parentObjType = self.convertFromStr(self.getAttribute(node, 'parentObjType'), 'str')
        
        data = None
        
        # read children
        for child in list(node.childNodes):
            if child.nodeName == 'module':
                data = self.getDao('module').fromXML(child)
            elif child.nodeName == 'location':
                data = self.getDao('location').fromXML(child)
            elif child.nodeName == 'annotation':
                data = self.getDao('annotation').fromXML(child)
            elif child.nodeName == 'function':
                data = self.getDao('function').fromXML(child)
            elif child.nodeName == 'connection':
                data = self.getDao('connection').fromXML(child)
            elif child.nodeName == 'port':
                data = self.getDao('port').fromXML(child)
            elif child.nodeName == 'parameter':
                data = self.getDao('parameter').fromXML(child)
            elif child.nodeName == 'portSpec':
                data = self.getDao('portSpec').fromXML(child)
            elif child.nodeName == 'abstractionRef':
                data = self.getDao('abstractionRef').fromXML(child)
            elif child.nodeName == 'other':
                data = self.getDao('other').fromXML(child)
            elif child.nodeType == child.TEXT_NODE and child.nodeValue.strip() == '':
                # node.removeChild(child)
                pass
            elif child.nodeType != child.TEXT_NODE:
                print '*** ERROR *** nodeName = %s' % child.nodeName
        
        obj = DBAdd(id=id,
                    what=what,
                    objectId=objectId,
                    parentObjId=parentObjId,
                    parentObjType=parentObjType,
                    data=data)
        obj.is_dirty = False
        return obj
    
    def toXML(self, add, doc, node):
        if node is None:
            node = doc.createElement('add')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(add.db_id, 'long'))
        node.setAttribute('what',self.convertToStr(add.db_what, 'str'))
        node.setAttribute('objectId',self.convertToStr(add.db_objectId, 'long'))
        node.setAttribute('parentObjId',self.convertToStr(add.db_parentObjId, 'long'))
        node.setAttribute('parentObjType',self.convertToStr(add.db_parentObjType, 'str'))
        
        # load DOM node map
        nodeMap = {}
        for childNode in node.childNodes:
            if childNode.nodeType == childNode.ELEMENT_NODE and self.hasAttribute(childNode, 'id'):
                nodeId = self.convertFromStr(self.getAttribute(childNode, 'id'), 'long')
                nodeMap[(childNode.nodeName, nodeId)] = childNode
        
        # set elements
        data = add.db_data
        if data.vtType == 'module':
            if nodeMap.has_key(('module', data.db_id)):
                childNode = nodeMap[('module', data.db_id)]
                del nodeMap[('module', data.db_id)]
            else:
                childNode = doc.createElement('module')
                node.appendChild(childNode)
            self.getDao('module').toXML(data, doc, childNode)
        elif data.vtType == 'location':
            if nodeMap.has_key(('location', data.db_id)):
                childNode = nodeMap[('location', data.db_id)]
                del nodeMap[('location', data.db_id)]
            else:
                childNode = doc.createElement('location')
                node.appendChild(childNode)
            self.getDao('location').toXML(data, doc, childNode)
        elif data.vtType == 'annotation':
            if nodeMap.has_key(('annotation', data.db_id)):
                childNode = nodeMap[('annotation', data.db_id)]
                del nodeMap[('annotation', data.db_id)]
            else:
                childNode = doc.createElement('annotation')
                node.appendChild(childNode)
            self.getDao('annotation').toXML(data, doc, childNode)
        elif data.vtType == 'function':
            if nodeMap.has_key(('function', data.db_id)):
                childNode = nodeMap[('function', data.db_id)]
                del nodeMap[('function', data.db_id)]
            else:
                childNode = doc.createElement('function')
                node.appendChild(childNode)
            self.getDao('function').toXML(data, doc, childNode)
        elif data.vtType == 'connection':
            if nodeMap.has_key(('connection', data.db_id)):
                childNode = nodeMap[('connection', data.db_id)]
                del nodeMap[('connection', data.db_id)]
            else:
                childNode = doc.createElement('connection')
                node.appendChild(childNode)
            self.getDao('connection').toXML(data, doc, childNode)
        elif data.vtType == 'port':
            if nodeMap.has_key(('port', data.db_id)):
                childNode = nodeMap[('port', data.db_id)]
                del nodeMap[('port', data.db_id)]
            else:
                childNode = doc.createElement('port')
                node.appendChild(childNode)
            self.getDao('port').toXML(data, doc, childNode)
        elif data.vtType == 'parameter':
            if nodeMap.has_key(('parameter', data.db_id)):
                childNode = nodeMap[('parameter', data.db_id)]
                del nodeMap[('parameter', data.db_id)]
            else:
                childNode = doc.createElement('parameter')
                node.appendChild(childNode)
            self.getDao('parameter').toXML(data, doc, childNode)
        elif data.vtType == 'portSpec':
            if nodeMap.has_key(('portSpec', data.db_id)):
                childNode = nodeMap[('portSpec', data.db_id)]
                del nodeMap[('portSpec', data.db_id)]
            else:
                childNode = doc.createElement('portSpec')
                node.appendChild(childNode)
            self.getDao('portSpec').toXML(data, doc, childNode)
        elif data.vtType == 'abstractionRef':
            if nodeMap.has_key(('abstractionRef', data.db_id)):
                childNode = nodeMap[('abstractionRef', data.db_id)]
                del nodeMap[('abstractionRef', data.db_id)]
            else:
                childNode = doc.createElement('abstractionRef')
                node.appendChild(childNode)
            self.getDao('abstractionRef').toXML(data, doc, childNode)
        elif data.vtType == 'other':
            if nodeMap.has_key(('other', data.db_id)):
                childNode = nodeMap[('other', data.db_id)]
                del nodeMap[('other', data.db_id)]
            else:
                childNode = doc.createElement('other')
                node.appendChild(childNode)
            self.getDao('other').toXML(data, doc, childNode)
        
        # delete nodes not around anymore
        for childNode in nodeMap.itervalues():
            childNode.parentNode.removeChild(childNode)
        return node

class DBOtherXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'other':
            return None
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        key = self.convertFromStr(self.getAttribute(node, 'key'), 'str')
        
        value = None
        
        # read children
        for child in list(node.childNodes):
            if child.nodeName == 'value':
                value = self.convertFromStr(child.firstChild.nodeValue,'str')
            elif child.nodeType == child.TEXT_NODE and child.nodeValue.strip() == '':
                # node.removeChild(child)
                pass
            elif child.nodeType != child.TEXT_NODE:
                print '*** ERROR *** nodeName = %s' % child.nodeName
        
        obj = DBOther(id=id,
                      key=key,
                      value=value)
        obj.is_dirty = False
        return obj
    
    def toXML(self, other, doc, node):
        if node is None:
            node = doc.createElement('other')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(other.db_id, 'long'))
        node.setAttribute('key',self.convertToStr(other.db_key, 'str'))
        
        # load DOM node map
        nodeMap = {}
        for childNode in node.childNodes:
            if childNode.nodeType == childNode.ELEMENT_NODE and self.hasAttribute(childNode, 'id'):
                nodeId = self.convertFromStr(self.getAttribute(childNode, 'id'), 'long')
                nodeMap[(childNode.nodeName, nodeId)] = childNode
        
        # set elements
        if other.db_value is not None:
            child = other.db_value
            if nodeMap.has_key(('value',child.db_id)):
                childNode = nodeMap[('value',child.db_id)]
                del nodeMap[('value', child.db_id)]
                textNode = childNode.firstChild
                textNode.replaceWholeText(self.convertToStr(child, 'str'))
            else:
                childNode = doc.createElement('value')
                node.appendChild(childNode)
                textNode = doc.createTextNode(self.convertToStr(child, 'str'))
                childNode.appendChild(textNode)
        
        # delete nodes not around anymore
        for childNode in nodeMap.itervalues():
            childNode.parentNode.removeChild(childNode)
        return node

class DBLocationXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'location':
            return None
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        x = self.convertFromStr(self.getAttribute(node, 'x'), 'float')
        y = self.convertFromStr(self.getAttribute(node, 'y'), 'float')
        
        obj = DBLocation(id=id,
                         x=x,
                         y=y)
        obj.is_dirty = False
        return obj
    
    def toXML(self, location, doc, node):
        if node is None:
            node = doc.createElement('location')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(location.db_id, 'long'))
        node.setAttribute('x',self.convertToStr(location.db_x, 'float'))
        node.setAttribute('y',self.convertToStr(location.db_y, 'float'))
        
        return node

class DBWorkflowExecXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'workflowExec':
            return None
        
        module_execs = {}
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        user = self.convertFromStr(self.getAttribute(node, 'user'), 'str')
        ip = self.convertFromStr(self.getAttribute(node, 'ip'), 'str')
        vt_version = self.convertFromStr(self.getAttribute(node, 'vtVersion'), 'str')
        ts_start = self.convertFromStr(self.getAttribute(node, 'tsStart'), 'datetime')
        ts_end = self.convertFromStr(self.getAttribute(node, 'tsEnd'), 'datetime')
        parent_id = self.convertFromStr(self.getAttribute(node, 'parentId'), 'long')
        parent_type = self.convertFromStr(self.getAttribute(node, 'parentType'), 'str')
        parent_version = self.convertFromStr(self.getAttribute(node, 'parentVersion'), 'long')
        name = self.convertFromStr(self.getAttribute(node, 'name'), 'str')
        
        
        # read children
        for child in list(node.childNodes):
            if child.nodeName == 'moduleExec':
                module_exec = self.getDao('module_exec').fromXML(child)
                module_execs[module_exec.db_id] = module_exec
            elif child.nodeType == child.TEXT_NODE and child.nodeValue.strip() == '':
                # node.removeChild(child)
                pass
            elif child.nodeType != child.TEXT_NODE:
                print '*** ERROR *** nodeName = %s' % child.nodeName
        
        obj = DBWorkflowExec(id=id,
                             user=user,
                             ip=ip,
                             vt_version=vt_version,
                             ts_start=ts_start,
                             ts_end=ts_end,
                             parent_id=parent_id,
                             parent_type=parent_type,
                             parent_version=parent_version,
                             name=name,
                             module_execs=module_execs)
        obj.is_dirty = False
        return obj
    
    def toXML(self, workflow_exec, doc, node):
        if node is None:
            node = doc.createElement('workflowExec')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(workflow_exec.db_id, 'long'))
        node.setAttribute('user',self.convertToStr(workflow_exec.db_user, 'str'))
        node.setAttribute('ip',self.convertToStr(workflow_exec.db_ip, 'str'))
        node.setAttribute('vtVersion',self.convertToStr(workflow_exec.db_vt_version, 'str'))
        node.setAttribute('tsStart',self.convertToStr(workflow_exec.db_ts_start, 'datetime'))
        node.setAttribute('tsEnd',self.convertToStr(workflow_exec.db_ts_end, 'datetime'))
        node.setAttribute('parentId',self.convertToStr(workflow_exec.db_parent_id, 'long'))
        node.setAttribute('parentType',self.convertToStr(workflow_exec.db_parent_type, 'str'))
        node.setAttribute('parentVersion',self.convertToStr(workflow_exec.db_parent_version, 'long'))
        node.setAttribute('name',self.convertToStr(workflow_exec.db_name, 'str'))
        
        # load DOM node map
        nodeMap = {}
        for childNode in node.childNodes:
            if childNode.nodeType == childNode.ELEMENT_NODE and self.hasAttribute(childNode, 'id'):
                nodeId = self.convertFromStr(self.getAttribute(childNode, 'id'), 'long')
                nodeMap[(childNode.nodeName, nodeId)] = childNode
        
        # set elements
        module_execs = workflow_exec.db_module_execs
        for module_exec in module_execs.itervalues():
            if nodeMap.has_key(('module_exec', module_exec.db_id)):
                childNode = nodeMap[('module_exec', module_exec.db_id)]
                del nodeMap[('module_exec', module_exec.db_id)]
            else:
                childNode = doc.createElement('module_exec')
                node.appendChild(childNode)
            self.getDao('module_exec').toXML(module_exec, doc, childNode)
        
        # delete nodes not around anymore
        for childNode in nodeMap.itervalues():
            childNode.parentNode.removeChild(childNode)
        return node

class DBFunctionXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'function':
            return None
        
        parameters = []
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        pos = self.convertFromStr(self.getAttribute(node, 'pos'), 'long')
        name = self.convertFromStr(self.getAttribute(node, 'name'), 'str')
        
        
        # read children
        for child in list(node.childNodes):
            if child.nodeName == 'parameter':
                parameter = self.getDao('parameter').fromXML(child)
                parameters.append(parameter)
            elif child.nodeType == child.TEXT_NODE and child.nodeValue.strip() == '':
                # node.removeChild(child)
                pass
            elif child.nodeType != child.TEXT_NODE:
                print '*** ERROR *** nodeName = %s' % child.nodeName
        
        obj = DBFunction(id=id,
                         pos=pos,
                         name=name,
                         parameters=parameters)
        obj.is_dirty = False
        return obj
    
    def toXML(self, function, doc, node):
        if node is None:
            node = doc.createElement('function')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(function.db_id, 'long'))
        node.setAttribute('pos',self.convertToStr(function.db_pos, 'long'))
        node.setAttribute('name',self.convertToStr(function.db_name, 'str'))
        
        # load DOM node map
        nodeMap = {}
        for childNode in node.childNodes:
            if childNode.nodeType == childNode.ELEMENT_NODE and self.hasAttribute(childNode, 'id'):
                nodeId = self.convertFromStr(self.getAttribute(childNode, 'id'), 'long')
                nodeMap[(childNode.nodeName, nodeId)] = childNode
        
        # set elements
        parameters = function.db_parameters
        for parameter in parameters:
            if nodeMap.has_key(('parameter', parameter.db_id)):
                childNode = nodeMap[('parameter', parameter.db_id)]
                del nodeMap[('parameter', parameter.db_id)]
            else:
                childNode = doc.createElement('parameter')
                node.appendChild(childNode)
            self.getDao('parameter').toXML(parameter, doc, childNode)
        
        # delete nodes not around anymore
        for childNode in nodeMap.itervalues():
            childNode.parentNode.removeChild(childNode)
        return node

class DBAbstractionXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'abstraction':
            return None
        
        actions = {}
        tags = {}
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        name = self.convertFromStr(self.getAttribute(node, 'name'), 'str')
        
        
        # read children
        for child in list(node.childNodes):
            if child.nodeName == 'action':
                action = self.getDao('action').fromXML(child)
                actions[action.db_id] = action
            elif child.nodeName == 'tag':
                tag = self.getDao('tag').fromXML(child)
                tags[tag.db_id] = tag
            elif child.nodeType == child.TEXT_NODE and child.nodeValue.strip() == '':
                # node.removeChild(child)
                pass
            elif child.nodeType != child.TEXT_NODE:
                print '*** ERROR *** nodeName = %s' % child.nodeName
        
        obj = DBAbstraction(id=id,
                            name=name,
                            actions=actions,
                            tags=tags)
        obj.is_dirty = False
        return obj
    
    def toXML(self, abstraction, doc, node):
        if node is None:
            node = doc.createElement('abstraction')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(abstraction.db_id, 'long'))
        node.setAttribute('name',self.convertToStr(abstraction.db_name, 'str'))
        
        # load DOM node map
        nodeMap = {}
        for childNode in node.childNodes:
            if childNode.nodeType == childNode.ELEMENT_NODE and self.hasAttribute(childNode, 'id'):
                nodeId = self.convertFromStr(self.getAttribute(childNode, 'id'), 'long')
                nodeMap[(childNode.nodeName, nodeId)] = childNode
        
        # set elements
        actions = abstraction.db_actions
        for action in actions.itervalues():
            if nodeMap.has_key(('action', action.db_id)):
                childNode = nodeMap[('action', action.db_id)]
                del nodeMap[('action', action.db_id)]
            else:
                childNode = doc.createElement('action')
                node.appendChild(childNode)
            self.getDao('action').toXML(action, doc, childNode)
        tags = abstraction.db_tags
        for tag in tags.itervalues():
            if nodeMap.has_key(('tag', tag.db_id)):
                childNode = nodeMap[('tag', tag.db_id)]
                del nodeMap[('tag', tag.db_id)]
            else:
                childNode = doc.createElement('tag')
                node.appendChild(childNode)
            self.getDao('tag').toXML(tag, doc, childNode)
        
        # delete nodes not around anymore
        for childNode in nodeMap.itervalues():
            childNode.parentNode.removeChild(childNode)
        return node

class DBWorkflowXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'workflow':
            return None
        
        annotations = []
        others = []
        
        modules = {}
        connections = {}
        abstractionRefs = {}
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        name = self.convertFromStr(self.getAttribute(node, 'name'), 'str')
        
        
        # read children
        for child in list(node.childNodes):
            if child.nodeName == 'module':
                module = self.getDao('module').fromXML(child)
                modules[module.db_id] = module
            elif child.nodeName == 'connection':
                connection = self.getDao('connection').fromXML(child)
                connections[connection.db_id] = connection
            elif child.nodeName == 'annotation':
                annotation = self.getDao('annotation').fromXML(child)
                annotations.append(annotation)
            elif child.nodeName == 'other':
                other = self.getDao('other').fromXML(child)
                others.append(other)
            elif child.nodeName == 'abstractionRef':
                abstractionRef = self.getDao('abstractionRef').fromXML(child)
                abstractionRefs[abstractionRef.db_id] = abstractionRef
            elif child.nodeType == child.TEXT_NODE and child.nodeValue.strip() == '':
                # node.removeChild(child)
                pass
            elif child.nodeType != child.TEXT_NODE:
                print '*** ERROR *** nodeName = %s' % child.nodeName
        
        obj = DBWorkflow(id=id,
                         name=name,
                         modules=modules,
                         connections=connections,
                         annotations=annotations,
                         others=others,
                         abstractionRefs=abstractionRefs)
        obj.is_dirty = False
        return obj
    
    def toXML(self, workflow, doc, node):
        if node is None:
            node = doc.createElement('workflow')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(workflow.db_id, 'long'))
        node.setAttribute('name',self.convertToStr(workflow.db_name, 'str'))
        
        # load DOM node map
        nodeMap = {}
        for childNode in node.childNodes:
            if childNode.nodeType == childNode.ELEMENT_NODE and self.hasAttribute(childNode, 'id'):
                nodeId = self.convertFromStr(self.getAttribute(childNode, 'id'), 'long')
                nodeMap[(childNode.nodeName, nodeId)] = childNode
        
        # set elements
        modules = workflow.db_modules
        for module in modules.itervalues():
            if nodeMap.has_key(('module', module.db_id)):
                childNode = nodeMap[('module', module.db_id)]
                del nodeMap[('module', module.db_id)]
            else:
                childNode = doc.createElement('module')
                node.appendChild(childNode)
            self.getDao('module').toXML(module, doc, childNode)
        connections = workflow.db_connections
        for connection in connections.itervalues():
            if nodeMap.has_key(('connection', connection.db_id)):
                childNode = nodeMap[('connection', connection.db_id)]
                del nodeMap[('connection', connection.db_id)]
            else:
                childNode = doc.createElement('connection')
                node.appendChild(childNode)
            self.getDao('connection').toXML(connection, doc, childNode)
        annotations = workflow.db_annotations
        for annotation in annotations:
            if nodeMap.has_key(('annotation', annotation.db_id)):
                childNode = nodeMap[('annotation', annotation.db_id)]
                del nodeMap[('annotation', annotation.db_id)]
            else:
                childNode = doc.createElement('annotation')
                node.appendChild(childNode)
            self.getDao('annotation').toXML(annotation, doc, childNode)
        others = workflow.db_others
        for other in others:
            if nodeMap.has_key(('other', other.db_id)):
                childNode = nodeMap[('other', other.db_id)]
                del nodeMap[('other', other.db_id)]
            else:
                childNode = doc.createElement('other')
                node.appendChild(childNode)
            self.getDao('other').toXML(other, doc, childNode)
        abstractionRefs = workflow.db_abstractionRefs
        for abstractionRef in abstractionRefs.itervalues():
            if nodeMap.has_key(('abstractionRef', abstractionRef.db_id)):
                childNode = nodeMap[('abstractionRef', abstractionRef.db_id)]
                del nodeMap[('abstractionRef', abstractionRef.db_id)]
            else:
                childNode = doc.createElement('abstractionRef')
                node.appendChild(childNode)
            self.getDao('abstractionRef').toXML(abstractionRef, doc, childNode)
        
        # delete nodes not around anymore
        for childNode in nodeMap.itervalues():
            childNode.parentNode.removeChild(childNode)
        return node

class DBAbstractionRefXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'abstractionRef':
            return None
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        abstraction_id = self.convertFromStr(self.getAttribute(node, 'abstractionId'), 'long')
        version = self.convertFromStr(self.getAttribute(node, 'version'), 'long')
        
        obj = DBAbstractionRef(id=id,
                               abstraction_id=abstraction_id,
                               version=version)
        obj.is_dirty = False
        return obj
    
    def toXML(self, abstractionRef, doc, node):
        if node is None:
            node = doc.createElement('abstractionRef')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(abstractionRef.db_id, 'long'))
        node.setAttribute('abstractionId',self.convertToStr(abstractionRef.db_abstraction_id, 'long'))
        node.setAttribute('version',self.convertToStr(abstractionRef.db_version, 'long'))
        
        return node

class DBAnnotationXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'annotation':
            return None
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        key = self.convertFromStr(self.getAttribute(node, 'key'), 'str')
        value = self.convertFromStr(self.getAttribute(node, 'value'), 'str')
        
        obj = DBAnnotation(id=id,
                           key=key,
                           value=value)
        obj.is_dirty = False
        return obj
    
    def toXML(self, annotation, doc, node):
        if node is None:
            node = doc.createElement('annotation')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(annotation.db_id, 'long'))
        node.setAttribute('key',self.convertToStr(annotation.db_key, 'str'))
        node.setAttribute('value',self.convertToStr(annotation.db_value, 'str'))
        
        return node

class DBChangeXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'change':
            return None
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        what = self.convertFromStr(self.getAttribute(node, 'what'), 'str')
        oldObjId = self.convertFromStr(self.getAttribute(node, 'oldObjId'), 'long')
        newObjId = self.convertFromStr(self.getAttribute(node, 'newObjId'), 'long')
        parentObjId = self.convertFromStr(self.getAttribute(node, 'parentObjId'), 'long')
        parentObjType = self.convertFromStr(self.getAttribute(node, 'parentObjType'), 'str')
        
        data = None
        
        # read children
        for child in list(node.childNodes):
            if child.nodeName == 'module':
                data = self.getDao('module').fromXML(child)
            elif child.nodeName == 'location':
                data = self.getDao('location').fromXML(child)
            elif child.nodeName == 'annotation':
                data = self.getDao('annotation').fromXML(child)
            elif child.nodeName == 'function':
                data = self.getDao('function').fromXML(child)
            elif child.nodeName == 'connection':
                data = self.getDao('connection').fromXML(child)
            elif child.nodeName == 'port':
                data = self.getDao('port').fromXML(child)
            elif child.nodeName == 'parameter':
                data = self.getDao('parameter').fromXML(child)
            elif child.nodeName == 'portSpec':
                data = self.getDao('portSpec').fromXML(child)
            elif child.nodeName == 'abstractionRef':
                data = self.getDao('abstractionRef').fromXML(child)
            elif child.nodeName == 'other':
                data = self.getDao('other').fromXML(child)
            elif child.nodeType == child.TEXT_NODE and child.nodeValue.strip() == '':
                # node.removeChild(child)
                pass
            elif child.nodeType != child.TEXT_NODE:
                print '*** ERROR *** nodeName = %s' % child.nodeName
        
        obj = DBChange(id=id,
                       what=what,
                       oldObjId=oldObjId,
                       newObjId=newObjId,
                       parentObjId=parentObjId,
                       parentObjType=parentObjType,
                       data=data)
        obj.is_dirty = False
        return obj
    
    def toXML(self, change, doc, node):
        if node is None:
            node = doc.createElement('change')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(change.db_id, 'long'))
        node.setAttribute('what',self.convertToStr(change.db_what, 'str'))
        node.setAttribute('oldObjId',self.convertToStr(change.db_oldObjId, 'long'))
        node.setAttribute('newObjId',self.convertToStr(change.db_newObjId, 'long'))
        node.setAttribute('parentObjId',self.convertToStr(change.db_parentObjId, 'long'))
        node.setAttribute('parentObjType',self.convertToStr(change.db_parentObjType, 'str'))
        
        # load DOM node map
        nodeMap = {}
        for childNode in node.childNodes:
            if childNode.nodeType == childNode.ELEMENT_NODE and self.hasAttribute(childNode, 'id'):
                nodeId = self.convertFromStr(self.getAttribute(childNode, 'id'), 'long')
                nodeMap[(childNode.nodeName, nodeId)] = childNode
        
        # set elements
        data = change.db_data
        if data.vtType == 'module':
            if nodeMap.has_key(('module', data.db_id)):
                childNode = nodeMap[('module', data.db_id)]
                del nodeMap[('module', data.db_id)]
            else:
                childNode = doc.createElement('module')
                node.appendChild(childNode)
            self.getDao('module').toXML(data, doc, childNode)
        elif data.vtType == 'location':
            if nodeMap.has_key(('location', data.db_id)):
                childNode = nodeMap[('location', data.db_id)]
                del nodeMap[('location', data.db_id)]
            else:
                childNode = doc.createElement('location')
                node.appendChild(childNode)
            self.getDao('location').toXML(data, doc, childNode)
        elif data.vtType == 'annotation':
            if nodeMap.has_key(('annotation', data.db_id)):
                childNode = nodeMap[('annotation', data.db_id)]
                del nodeMap[('annotation', data.db_id)]
            else:
                childNode = doc.createElement('annotation')
                node.appendChild(childNode)
            self.getDao('annotation').toXML(data, doc, childNode)
        elif data.vtType == 'function':
            if nodeMap.has_key(('function', data.db_id)):
                childNode = nodeMap[('function', data.db_id)]
                del nodeMap[('function', data.db_id)]
            else:
                childNode = doc.createElement('function')
                node.appendChild(childNode)
            self.getDao('function').toXML(data, doc, childNode)
        elif data.vtType == 'connection':
            if nodeMap.has_key(('connection', data.db_id)):
                childNode = nodeMap[('connection', data.db_id)]
                del nodeMap[('connection', data.db_id)]
            else:
                childNode = doc.createElement('connection')
                node.appendChild(childNode)
            self.getDao('connection').toXML(data, doc, childNode)
        elif data.vtType == 'port':
            if nodeMap.has_key(('port', data.db_id)):
                childNode = nodeMap[('port', data.db_id)]
                del nodeMap[('port', data.db_id)]
            else:
                childNode = doc.createElement('port')
                node.appendChild(childNode)
            self.getDao('port').toXML(data, doc, childNode)
        elif data.vtType == 'parameter':
            if nodeMap.has_key(('parameter', data.db_id)):
                childNode = nodeMap[('parameter', data.db_id)]
                del nodeMap[('parameter', data.db_id)]
            else:
                childNode = doc.createElement('parameter')
                node.appendChild(childNode)
            self.getDao('parameter').toXML(data, doc, childNode)
        elif data.vtType == 'portSpec':
            if nodeMap.has_key(('portSpec', data.db_id)):
                childNode = nodeMap[('portSpec', data.db_id)]
                del nodeMap[('portSpec', data.db_id)]
            else:
                childNode = doc.createElement('portSpec')
                node.appendChild(childNode)
            self.getDao('portSpec').toXML(data, doc, childNode)
        elif data.vtType == 'abstractionRef':
            if nodeMap.has_key(('abstractionRef', data.db_id)):
                childNode = nodeMap[('abstractionRef', data.db_id)]
                del nodeMap[('abstractionRef', data.db_id)]
            else:
                childNode = doc.createElement('abstractionRef')
                node.appendChild(childNode)
            self.getDao('abstractionRef').toXML(data, doc, childNode)
        elif data.vtType == 'other':
            if nodeMap.has_key(('other', data.db_id)):
                childNode = nodeMap[('other', data.db_id)]
                del nodeMap[('other', data.db_id)]
            else:
                childNode = doc.createElement('other')
                node.appendChild(childNode)
            self.getDao('other').toXML(data, doc, childNode)
        
        # delete nodes not around anymore
        for childNode in nodeMap.itervalues():
            childNode.parentNode.removeChild(childNode)
        return node

class DBParameterXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'parameter':
            return None
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        pos = self.convertFromStr(self.getAttribute(node, 'pos'), 'long')
        name = self.convertFromStr(self.getAttribute(node, 'name'), 'str')
        type = self.convertFromStr(self.getAttribute(node, 'type'), 'str')
        val = self.convertFromStr(self.getAttribute(node, 'val'), 'str')
        alias = self.convertFromStr(self.getAttribute(node, 'alias'), 'str')
        
        obj = DBParameter(id=id,
                          pos=pos,
                          name=name,
                          type=type,
                          val=val,
                          alias=alias)
        obj.is_dirty = False
        return obj
    
    def toXML(self, parameter, doc, node):
        if node is None:
            node = doc.createElement('parameter')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(parameter.db_id, 'long'))
        node.setAttribute('pos',self.convertToStr(parameter.db_pos, 'long'))
        node.setAttribute('name',self.convertToStr(parameter.db_name, 'str'))
        node.setAttribute('type',self.convertToStr(parameter.db_type, 'str'))
        node.setAttribute('val',self.convertToStr(parameter.db_val, 'str'))
        node.setAttribute('alias',self.convertToStr(parameter.db_alias, 'str'))
        
        return node

class DBConnectionXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'connection':
            return None
        
        ports = []
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        
        
        # read children
        for child in list(node.childNodes):
            if child.nodeName == 'port':
                port = self.getDao('port').fromXML(child)
                ports.append(port)
            elif child.nodeType == child.TEXT_NODE and child.nodeValue.strip() == '':
                # node.removeChild(child)
                pass
            elif child.nodeType != child.TEXT_NODE:
                print '*** ERROR *** nodeName = %s' % child.nodeName
        
        obj = DBConnection(id=id,
                           ports=ports)
        obj.is_dirty = False
        return obj
    
    def toXML(self, connection, doc, node):
        if node is None:
            node = doc.createElement('connection')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(connection.db_id, 'long'))
        
        # load DOM node map
        nodeMap = {}
        for childNode in node.childNodes:
            if childNode.nodeType == childNode.ELEMENT_NODE and self.hasAttribute(childNode, 'id'):
                nodeId = self.convertFromStr(self.getAttribute(childNode, 'id'), 'long')
                nodeMap[(childNode.nodeName, nodeId)] = childNode
        
        # set elements
        ports = connection.db_ports
        for port in ports:
            if nodeMap.has_key(('port', port.db_id)):
                childNode = nodeMap[('port', port.db_id)]
                del nodeMap[('port', port.db_id)]
            else:
                childNode = doc.createElement('port')
                node.appendChild(childNode)
            self.getDao('port').toXML(port, doc, childNode)
        
        # delete nodes not around anymore
        for childNode in nodeMap.itervalues():
            childNode.parentNode.removeChild(childNode)
        return node

class DBActionXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'action':
            return None
        
        operations = []
        
        annotations = {}
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        prevId = self.convertFromStr(self.getAttribute(node, 'prevId'), 'long')
        date = self.convertFromStr(self.getAttribute(node, 'date'), 'datetime')
        user = self.convertFromStr(self.getAttribute(node, 'user'), 'str')
        
        
        # read children
        for child in list(node.childNodes):
            if child.nodeName == 'annotation':
                annotation = self.getDao('annotation').fromXML(child)
                annotations[annotation.db_id] = annotation
            elif child.nodeName == 'add':
                operation = self.getDao('add').fromXML(child)
                operations.append(operation)
            elif child.nodeName == 'delete':
                operation = self.getDao('delete').fromXML(child)
                operations.append(operation)
            elif child.nodeName == 'change':
                operation = self.getDao('change').fromXML(child)
                operations.append(operation)
            elif child.nodeType == child.TEXT_NODE and child.nodeValue.strip() == '':
                # node.removeChild(child)
                pass
            elif child.nodeType != child.TEXT_NODE:
                print '*** ERROR *** nodeName = %s' % child.nodeName
        
        obj = DBAction(id=id,
                       prevId=prevId,
                       date=date,
                       user=user,
                       annotations=annotations,
                       operations=operations)
        obj.is_dirty = False
        return obj
    
    def toXML(self, action, doc, node):
        if node is None:
            node = doc.createElement('action')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(action.db_id, 'long'))
        node.setAttribute('prevId',self.convertToStr(action.db_prevId, 'long'))
        node.setAttribute('date',self.convertToStr(action.db_date, 'datetime'))
        node.setAttribute('user',self.convertToStr(action.db_user, 'str'))
        
        # load DOM node map
        nodeMap = {}
        for childNode in node.childNodes:
            if childNode.nodeType == childNode.ELEMENT_NODE and self.hasAttribute(childNode, 'id'):
                nodeId = self.convertFromStr(self.getAttribute(childNode, 'id'), 'long')
                nodeMap[(childNode.nodeName, nodeId)] = childNode
        
        # set elements
        annotations = action.db_annotations
        for annotation in annotations.itervalues():
            if nodeMap.has_key(('annotation', annotation.db_id)):
                childNode = nodeMap[('annotation', annotation.db_id)]
                del nodeMap[('annotation', annotation.db_id)]
            else:
                childNode = doc.createElement('annotation')
                node.appendChild(childNode)
            self.getDao('annotation').toXML(annotation, doc, childNode)
        operations = action.db_operations
        for operation in operations:
            if operation.vtType == 'add':
                if nodeMap.has_key(('add', operation.db_id)):
                    childNode = nodeMap[('add', operation.db_id)]
                    del nodeMap[('add', operation.db_id)]
                else:
                    childNode = doc.createElement('add')
                    node.appendChild(childNode)
                self.getDao('add').toXML(operation, doc, childNode)
            elif operation.vtType == 'delete':
                if nodeMap.has_key(('delete', operation.db_id)):
                    childNode = nodeMap[('delete', operation.db_id)]
                    del nodeMap[('delete', operation.db_id)]
                else:
                    childNode = doc.createElement('delete')
                    node.appendChild(childNode)
                self.getDao('delete').toXML(operation, doc, childNode)
            elif operation.vtType == 'change':
                if nodeMap.has_key(('change', operation.db_id)):
                    childNode = nodeMap[('change', operation.db_id)]
                    del nodeMap[('change', operation.db_id)]
                else:
                    childNode = doc.createElement('change')
                    node.appendChild(childNode)
                self.getDao('change').toXML(operation, doc, childNode)
        
        # delete nodes not around anymore
        for childNode in nodeMap.itervalues():
            childNode.parentNode.removeChild(childNode)
        return node

class DBDeleteXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'delete':
            return None
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        what = self.convertFromStr(self.getAttribute(node, 'what'), 'str')
        objectId = self.convertFromStr(self.getAttribute(node, 'objectId'), 'long')
        parentObjId = self.convertFromStr(self.getAttribute(node, 'parentObjId'), 'long')
        parentObjType = self.convertFromStr(self.getAttribute(node, 'parentObjType'), 'str')
        
        obj = DBDelete(id=id,
                       what=what,
                       objectId=objectId,
                       parentObjId=parentObjId,
                       parentObjType=parentObjType)
        obj.is_dirty = False
        return obj
    
    def toXML(self, delete, doc, node):
        if node is None:
            node = doc.createElement('delete')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(delete.db_id, 'long'))
        node.setAttribute('what',self.convertToStr(delete.db_what, 'str'))
        node.setAttribute('objectId',self.convertToStr(delete.db_objectId, 'long'))
        node.setAttribute('parentObjId',self.convertToStr(delete.db_parentObjId, 'long'))
        node.setAttribute('parentObjType',self.convertToStr(delete.db_parentObjType, 'str'))
        
        return node

class DBVistrailXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'vistrail':
            return None
        
        actions = {}
        tags = {}
        abstractions = {}
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        version = self.convertFromStr(self.getAttribute(node, 'version'), 'str')
        name = self.convertFromStr(self.getAttribute(node, 'name'), 'str')
        dbHost = self.convertFromStr(self.getAttribute(node, 'dbHost'), 'str')
        dbPort = self.convertFromStr(self.getAttribute(node, 'dbPort'), 'int')
        dbName = self.convertFromStr(self.getAttribute(node, 'dbName'), 'str')
        
        
        # read children
        for child in list(node.childNodes):
            if child.nodeName == 'action':
                action = self.getDao('action').fromXML(child)
                actions[action.db_id] = action
            elif child.nodeName == 'tag':
                tag = self.getDao('tag').fromXML(child)
                tags[tag.db_id] = tag
            elif child.nodeName == 'abstraction':
                abstraction = self.getDao('abstraction').fromXML(child)
                abstractions[abstraction.db_id] = abstraction
            elif child.nodeType == child.TEXT_NODE and child.nodeValue.strip() == '':
                # node.removeChild(child)
                pass
            elif child.nodeType != child.TEXT_NODE:
                print '*** ERROR *** nodeName = %s' % child.nodeName
        
        obj = DBVistrail(id=id,
                         version=version,
                         name=name,
                         dbHost=dbHost,
                         dbPort=dbPort,
                         dbName=dbName,
                         actions=actions,
                         tags=tags,
                         abstractions=abstractions)
        obj.is_dirty = False
        return obj
    
    def toXML(self, vistrail, doc, node):
        if node is None:
            node = doc.createElement('vistrail')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(vistrail.db_id, 'long'))
        node.setAttribute('version',self.convertToStr(vistrail.db_version, 'str'))
        node.setAttribute('name',self.convertToStr(vistrail.db_name, 'str'))
        node.setAttribute('dbHost',self.convertToStr(vistrail.db_dbHost, 'str'))
        node.setAttribute('dbPort',self.convertToStr(vistrail.db_dbPort, 'int'))
        node.setAttribute('dbName',self.convertToStr(vistrail.db_dbName, 'str'))
        
        # load DOM node map
        nodeMap = {}
        for childNode in node.childNodes:
            if childNode.nodeType == childNode.ELEMENT_NODE and self.hasAttribute(childNode, 'id'):
                nodeId = self.convertFromStr(self.getAttribute(childNode, 'id'), 'long')
                nodeMap[(childNode.nodeName, nodeId)] = childNode
        
        # set elements
        actions = vistrail.db_actions
        for action in actions.itervalues():
            if nodeMap.has_key(('action', action.db_id)):
                childNode = nodeMap[('action', action.db_id)]
                del nodeMap[('action', action.db_id)]
            else:
                childNode = doc.createElement('action')
                node.appendChild(childNode)
            self.getDao('action').toXML(action, doc, childNode)
        tags = vistrail.db_tags
        for tag in tags.itervalues():
            if nodeMap.has_key(('tag', tag.db_id)):
                childNode = nodeMap[('tag', tag.db_id)]
                del nodeMap[('tag', tag.db_id)]
            else:
                childNode = doc.createElement('tag')
                node.appendChild(childNode)
            self.getDao('tag').toXML(tag, doc, childNode)
        abstractions = vistrail.db_abstractions
        for abstraction in abstractions.itervalues():
            if nodeMap.has_key(('abstraction', abstraction.db_id)):
                childNode = nodeMap[('abstraction', abstraction.db_id)]
                del nodeMap[('abstraction', abstraction.db_id)]
            else:
                childNode = doc.createElement('abstraction')
                node.appendChild(childNode)
            self.getDao('abstraction').toXML(abstraction, doc, childNode)
        
        # delete nodes not around anymore
        for childNode in nodeMap.itervalues():
            childNode.parentNode.removeChild(childNode)
        return node

class DBModuleExecXMLDAOBase(XMLDAO):

    def __init__(self, daoList):
        self.daoList = daoList

    def getDao(self, dao):
        return self.daoList[dao]

    def fromXML(self, node):
        if node.nodeName != 'moduleExec':
            return None
        
        annotations = []
        
        # read attributes
        id = self.convertFromStr(self.getAttribute(node, 'id'), 'long')
        ts_start = self.convertFromStr(self.getAttribute(node, 'tsStart'), 'datetime')
        ts_end = self.convertFromStr(self.getAttribute(node, 'tsEnd'), 'datetime')
        module_id = self.convertFromStr(self.getAttribute(node, 'moduleId'), 'long')
        module_name = self.convertFromStr(self.getAttribute(node, 'moduleName'), 'str')
        
        
        # read children
        for child in list(node.childNodes):
            if child.nodeName == 'annotation':
                annotation = self.getDao('annotation').fromXML(child)
                annotations.append(annotation)
            elif child.nodeType == child.TEXT_NODE and child.nodeValue.strip() == '':
                # node.removeChild(child)
                pass
            elif child.nodeType != child.TEXT_NODE:
                print '*** ERROR *** nodeName = %s' % child.nodeName
        
        obj = DBModuleExec(id=id,
                           ts_start=ts_start,
                           ts_end=ts_end,
                           module_id=module_id,
                           module_name=module_name,
                           annotations=annotations)
        obj.is_dirty = False
        return obj
    
    def toXML(self, module_exec, doc, node):
        if node is None:
            node = doc.createElement('moduleExec')
        
        # set attributes
        node.setAttribute('id',self.convertToStr(module_exec.db_id, 'long'))
        node.setAttribute('tsStart',self.convertToStr(module_exec.db_ts_start, 'datetime'))
        node.setAttribute('tsEnd',self.convertToStr(module_exec.db_ts_end, 'datetime'))
        node.setAttribute('moduleId',self.convertToStr(module_exec.db_module_id, 'long'))
        node.setAttribute('moduleName',self.convertToStr(module_exec.db_module_name, 'str'))
        
        # load DOM node map
        nodeMap = {}
        for childNode in node.childNodes:
            if childNode.nodeType == childNode.ELEMENT_NODE and self.hasAttribute(childNode, 'id'):
                nodeId = self.convertFromStr(self.getAttribute(childNode, 'id'), 'long')
                nodeMap[(childNode.nodeName, nodeId)] = childNode
        
        # set elements
        annotations = module_exec.db_annotations
        for annotation in annotations:
            if nodeMap.has_key(('annotation', annotation.db_id)):
                childNode = nodeMap[('annotation', annotation.db_id)]
                del nodeMap[('annotation', annotation.db_id)]
            else:
                childNode = doc.createElement('annotation')
                node.appendChild(childNode)
            self.getDao('annotation').toXML(annotation, doc, childNode)
        
        # delete nodes not around anymore
        for childNode in nodeMap.itervalues():
            childNode.parentNode.removeChild(childNode)
        return node

"""generated automatically by auto_dao.py"""

class XMLDAOListBase(dict):

    def __init__(self, daos=None):
        if daos is not None:
            dict.update(self, daos)

        if 'portSpec' not in self:
            self['portSpec'] = DBPortSpecXMLDAOBase(self)
        if 'module' not in self:
            self['module'] = DBModuleXMLDAOBase(self)
        if 'tag' not in self:
            self['tag'] = DBTagXMLDAOBase(self)
        if 'port' not in self:
            self['port'] = DBPortXMLDAOBase(self)
        if 'log' not in self:
            self['log'] = DBLogXMLDAOBase(self)
        if 'machine' not in self:
            self['machine'] = DBMachineXMLDAOBase(self)
        if 'add' not in self:
            self['add'] = DBAddXMLDAOBase(self)
        if 'other' not in self:
            self['other'] = DBOtherXMLDAOBase(self)
        if 'location' not in self:
            self['location'] = DBLocationXMLDAOBase(self)
        if 'workflow_exec' not in self:
            self['workflow_exec'] = DBWorkflowExecXMLDAOBase(self)
        if 'function' not in self:
            self['function'] = DBFunctionXMLDAOBase(self)
        if 'abstraction' not in self:
            self['abstraction'] = DBAbstractionXMLDAOBase(self)
        if 'workflow' not in self:
            self['workflow'] = DBWorkflowXMLDAOBase(self)
        if 'abstractionRef' not in self:
            self['abstractionRef'] = DBAbstractionRefXMLDAOBase(self)
        if 'annotation' not in self:
            self['annotation'] = DBAnnotationXMLDAOBase(self)
        if 'change' not in self:
            self['change'] = DBChangeXMLDAOBase(self)
        if 'parameter' not in self:
            self['parameter'] = DBParameterXMLDAOBase(self)
        if 'connection' not in self:
            self['connection'] = DBConnectionXMLDAOBase(self)
        if 'action' not in self:
            self['action'] = DBActionXMLDAOBase(self)
        if 'delete' not in self:
            self['delete'] = DBDeleteXMLDAOBase(self)
        if 'vistrail' not in self:
            self['vistrail'] = DBVistrailXMLDAOBase(self)
        if 'module_exec' not in self:
            self['module_exec'] = DBModuleExecXMLDAOBase(self)



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################

from __future__ import division

import copy
import math
import os
import uuid

from PyQt4 import QtCore, QtGui

import vistrails.core.analogy
from vistrails.core.configuration import get_vistrails_configuration
from vistrails.core.data_structures.graph import Graph
from vistrails.core import debug
import vistrails.core.db.action
from vistrails.core.interpreter.default import get_default_interpreter
from vistrails.core.vistrail.job import Workflow as JobWorkflow
from vistrails.core.layout.version_tree_layout import VistrailsTreeLayoutLW
from vistrails.core.log.opm_graph import OpmGraph
from vistrails.core.log.prov_document import ProvDocument
from vistrails.core.modules.abstraction import identifier as abstraction_pkg
from vistrails.core.modules.module_registry import get_module_registry
from vistrails.core.param_explore import ActionBasedParameterExploration
from vistrails.core.query.version import TrueSearch
from vistrails.core.query.visual import VisualQuery
from vistrails.core.utils import DummyView, VistrailsInternalError, InvalidPipeline
import vistrails.core.system
from vistrails.core.vistrail.controller import VistrailController as BaseController
from vistrails.core.vistrail.pipeline import Pipeline
from vistrails.core.vistrail.vistrail import Vistrail, TagExists
from vistrails.gui.pipeline_view import QPipelineView
from vistrails.gui.theme import CurrentTheme
import vistrails.gui.utils
from vistrails.gui.utils import show_warning, show_question, YES_BUTTON, NO_BUTTON
from vistrails.gui.version_prop import QVersionProp



################################################################################

class ExecutionProgressDialog(QtGui.QProgressDialog):
    def __init__(self, parent=None):
        QtGui.QProgressDialog.__init__(self, 'Executing Workflow',
                                       '&Cancel',
                                       0, 100,
                                       parent, QtCore.Qt.Dialog)
        self.setWindowTitle('Executing')
        self.setWindowModality(QtCore.Qt.WindowModal)
        self._last_set_value = 0
        self._progress_canceled = False
        # if suspended is true we should not wait for a job to complete
        self.suspended = False

    def setValue(self, value):
        self._last_set_value = value
        super(ExecutionProgressDialog, self).setValue(value)

    def goOn(self):
        self.reset()
        self.show()
        super(ExecutionProgressDialog, self).setValue(self._last_set_value)


class VistrailController(QtCore.QObject, BaseController):
    """
    VistrailController is the class handling all action control in
    VisTrails. It updates pipeline, vistrail and emit signals to
    update the view

    Signals emitted:

    vistrailChanged(): emitted when the version tree needs to be
    recreated (for example, a node was added/deleted or the layout
    changed).

    versionWasChanged(): emitted when the current version (the one
    being displayed by the pipeline view) has changed.

    searchChanged(): emitted when the search statement from the
    version view has changed.

    stateChanged(): stateChanged is called when a vistrail goes from
    unsaved to saved or vice-versa.
    
    notesChanged(): notesChanged is called when the version notes have
    been updated

    """

    def __init__(self, vistrail=None, locator=None, abstractions=None,
                 thumbnails=None, mashups=None, pipeline_view=None, 
                 id_scope=None, set_log_on_vt=True, auto_save=True, name=''):
        """ VistrailController(vistrail: Vistrail, 
                               locator: BaseLocator,
                               abstractions: [<filename strings>],
                               thumbnails: [<filename strings>],
                               mashups: [<filename strings>],
                               pipeline_view: QPipelineView
                               id_scope: IdScope,
                               set_log_on_vt: bool,
                               auto_save: bool, 
                               name: str) -> VistrailController
        Create a controller for a vistrail.

        """

        QtCore.QObject.__init__(self)

        if pipeline_view is None:
            self.current_pipeline_view = QPipelineView()
        else:
            self.current_pipeline_view = pipeline_view

        self.vistrail_view = None
        self.reset_pipeline_view = False
        self.reset_version_view = True
        self.quiet = False
        self.progress = None
        self.create_job = False
        
        self.analogy = {}
        # if self._auto_save is True, an auto_saving timer will save a temporary
        # file every 2 minutes
        self._auto_save = auto_save
        self.timer = None
        if self._auto_save:
            self.setup_timer()

        def width_f(text):
            return CurrentTheme.VERSION_FONT_METRIC.width(text)
        self._current_graph_layout = \
            VistrailsTreeLayoutLW(width_f, 
                                  CurrentTheme.VERSION_FONT_METRIC.height(), 
                                  CurrentTheme.VERSION_LABEL_MARGIN[0], 
                                  CurrentTheme.VERSION_LABEL_MARGIN[1])
        #this was moved to BaseController
        #self.num_versions_always_shown = 1
        BaseController.__init__(self, vistrail, locator, abstractions, 
                                thumbnails, mashups, id_scope, set_log_on_vt, 
                                auto_save)

    def _get_current_pipeline_scene(self):
        return self.current_pipeline_view.scene()
    current_pipeline_scene = property(_get_current_pipeline_scene)

    # just need to switch current_pipeline_view to update controller to
    # new version and pipeline...
    def _get_current_version(self):
        if self.current_pipeline_view is None:
            return -1
        return self.current_pipeline_view.scene().current_version
    def _set_current_version(self, version):
        # print "set_current_version:", version, id(self.current_pipeline_view)
        if self.current_pipeline_view is not None:
            self.current_pipeline_view.scene().current_version = version
    current_version = property(_get_current_version, _set_current_version)

    def _get_current_pipeline(self):
        if self.current_pipeline_view is None:
            return None
        return self.current_pipeline_view.scene().current_pipeline
    def _set_current_pipeline(self, pipeline):
        if self.current_pipeline_view is not None:
            self.current_pipeline_view.scene().current_pipeline = pipeline
    current_pipeline = property(_get_current_pipeline, _set_current_pipeline)

    def set_pipeline_view(self, pipeline_view):
        if self.current_pipeline_view is not None:
            self.disconnect(self, QtCore.SIGNAL('versionWasChanged'),
                            self.current_pipeline_view.version_changed)
        self.current_pipeline_view = pipeline_view
        self.connect(self, QtCore.SIGNAL('versionWasChanged'),
                     self.current_pipeline_view.version_changed)
    
    def setup_timer(self):
        self.timer = QtCore.QTimer(self)
        self.connect(self.timer, QtCore.SIGNAL("timeout()"), self.write_temporary)
        self.timer.start(1000 * 60 * 2) # Save every two minutes
        
    def stop_timer(self):
        if self.timer:
            self.disconnect(self.timer, QtCore.SIGNAL("timeout()"), self.write_temporary)
            self.timer.stop()
            
    ##########################################################################
    # Signal vistrail relayout / redraw

    def replace_unnamed_node_in_version_tree(self, old_version, new_version):
        """method analogous to invalidate_version_tree but when only
        a single unnamed node and links need to be updated. Much faster."""
        self.reset_version_view = False
        try:
            self.emit(QtCore.SIGNAL('invalidateSingleNodeInVersionTree'),
                                    old_version, new_version)
        finally:
            self.reset_version_view = True

    def invalidate_version_tree(self, reset_version_view=True, animate_layout=False):
        """ invalidate_version_tree(reset_version_tree: bool, animate_layout: bool) -> None
        
        """
        self.reset_version_view = reset_version_view
        #FIXME: in the future, rename the signal
        try:
            self.emit(QtCore.SIGNAL('vistrailChanged()'))
        finally:
            self.reset_version_view = True

    def has_move_actions(self):
        return self.current_pipeline_scene.hasMoveActions()

    def flush_move_actions(self):
        return self.current_pipeline_scene.flushMoveActions()

    ##########################################################################
    # Pipeline View Methods

    def updatePipelineScene(self):
        self.current_pipeline_scene.setupScene(self.current_pipeline)

    ##########################################################################
    # Autosave

    def enable_autosave(self):
        self._auto_save = True

    def disable_autosave(self):
        self._auto_save = False

    def get_locator(self):
        from vistrails.gui.application import get_vistrails_application
        if (self._auto_save and 
            get_vistrails_application().configuration.check('autoSave')):
            if self.locator is None:
                raise ValueError("locator is None")
            return self.locator
        else:
            return None

    def cleanup(self):
        locator = self.get_locator()
        if locator:
            locator.clean_temporaries()
        if self._auto_save or self.timer:
            self.stop_timer()
        # close associated mashup apps
        version_prop = QVersionProp.instance()
        for app in version_prop.versionMashups.apps.values():
            if app and app.view == self.vistrail_view:
                app.close()


    ##########################################################################
    # Actions, etc
    
    def perform_action(self, action, quiet=None):
        """ performAction(action: Action, quiet=None) -> timestep

        performs given action on current pipeline.

        quiet and self.quiet control invalidation of version
        tree. If quiet is set to any value, it overrides the field
        value self.quiet.

        If the value is True, then no invalidation happens (gui is not
        updated.)
        
        """
        if action is not None:
            BaseController.perform_action(self,action)

            if quiet is None:
                if not self.quiet:
                    self.invalidate_version_tree(False)
            else:
                if not quiet:
                    self.invalidate_version_tree(False)
            return action.db_id
        return None

    def add_new_action(self, action, description=None):
        """add_new_action(action) -> None

        Call this function to add a new action to the vistrail being
        controlled by the vistrailcontroller.

        FIXME: In the future, this function should watch the vistrail
        and get notified of the change.

        """
        if action is not None:
            BaseController.add_new_action(self, action, description)
            self.emit(QtCore.SIGNAL("new_action"), action)
            self.recompute_terse_graph()

    ##########################################################################

    def add_module(self, x, y, identifier, name, namespace='', 
                   internal_version=-1):
        return BaseController.add_module(self, identifier, name, namespace, x, y,
                                         internal_version)

    def create_abstraction_with_prompt(self, module_ids, connection_ids, 
                                       name=""):
        name = self.get_abstraction_name(name)
        if name is None:
            return
        return self.create_abstraction(module_ids, connection_ids, name)

    def update_notes(self, notes):
        # Find current location of notes
        notes_version = self.vistrail.search_upgrade_versions(
                self.current_version,
                lambda vt, v, bv: v if vt.get_notes(v) else None)

        # Remove notes
        if notes_version is not None:
            self.vistrail.set_notes(notes_version, None)

        # Add notes
        if self.vistrail.set_notes(self.current_base_version, str(notes)):
            self.set_changed(True)
            self.emit(QtCore.SIGNAL('notesChanged()'))

    ##########################################################################
    # Workflow Execution
    
    def execute_workflow_list(self, vistrails):
        old_quiet = self.quiet
        self.quiet = True
        self.current_pipeline_scene.reset_module_colors()
        self.current_pipeline_scene.update()
        (results, changed) = BaseController.execute_workflow_list(self, 
                                                                  vistrails)        
        self.quiet = old_quiet
        if changed:
            self.invalidate_version_tree(False)
        return (results, changed)

    def execute_current_workflow(self, custom_aliases=None, custom_params=None,
                                 extra_info=None, reason='Pipeline Execution',
                                 sinks=None):
        """ execute_current_workflow() -> None
        Execute the current workflow (if exists)
        
        """
        self.flush_delayed_actions()

        if self.create_job:
            version_id = self.current_version
            # check if a job exist for this workflow
            current_workflow = None
            for wf in self.jobMonitor.workflows.itervalues():
                try:
                    wf_version = int(wf.version)
                except ValueError:
                    wf_version = self.vistrail.get_version_number(wf.version)
                if version_id == wf_version:
                    current_workflow = wf
                    self.jobMonitor.startWorkflow(wf)
            if not current_workflow:
                current_workflow = JobWorkflow(version_id)
                self.jobMonitor.startWorkflow(current_workflow)
            self.create_job = False

        if self.current_pipeline:
            locator = self.get_locator()
            if locator:
                locator.clean_temporaries()
                locator.save_temporary(self.vistrail)
            try:
                return self.execute_workflow_list([(self.locator,
                                             self.current_version,
                                             self.current_pipeline,
                                             self.current_pipeline_scene,
                                             custom_aliases,
                                             custom_params,
                                             reason,
                                             sinks,
                                             extra_info)])
            except Exception, e:
                debug.unexpected_exception(e)
                raise
        return ([], False)


    def execute_user_workflow(self, reason='Pipeline Execution', sinks=None):
        """ execute_user_workflow() -> None
        Execute the current workflow (if exists) and monitors it if it contains jobs
        
        """

        # reset job view
        from vistrails.gui.job_monitor import QJobView
        jobView = QJobView.instance()
        if jobView.updating_now:
            debug.critical("Execution Aborted: Job Monitor is updating. "
                           "Please wait a few seconds and try again.")
            return
        jobView.updating_now = True

        try:
            self.progress = ExecutionProgressDialog(self.vistrail_view)
            self.progress.show()

            if not self.jobMonitor.currentWorkflow():
                self.create_job = True

            result =  self.execute_current_workflow(reason=reason, sinks=sinks)

            self.progress.setValue(100)
        finally:
            self.progress.hide()
            self.progress.deleteLater()
            self.progress = None
            self.create_job = False
            self.jobMonitor.finishWorkflow()
            jobView.updating_now = False

        return result

    def enable_missing_package(self, identifier, deps):
        configuration = get_vistrails_configuration()
        if getattr(configuration, 'enablePackagesSilently', False):
            return True

        msg = "VisTrails needs to enable package '%s'." % identifier
        if len(deps) > 0:
            msg += (" This will also enable the dependencies: %s." 
                    " Do you want to enable these packages?" % (
                    ", ".join(deps),))
        else:
            msg += " Do you want to enable this package?"
        res = show_question('Enable package?',
                            msg,
                            [YES_BUTTON, NO_BUTTON], 
                            YES_BUTTON)
        if res == NO_BUTTON:
            return False
        return True

    def install_missing_package(self, identifier):
        res = show_question('Install package?',
                            "This pipeline contains a module"
                            " in package '%s', which"
                            " is not installed. Do you want to"
                            " install and enable that package?" % \
                                identifier, [YES_BUTTON, NO_BUTTON],
                            YES_BUTTON)
        return res == YES_BUTTON

    def change_selected_version(self, new_version, report_all_errors=True,
                                do_validate=True, from_root=False):
        """change_selected_version(new_version: int,
                                   report_all_errors: boolean,
                                   do_validate: boolean,
                                   from_root: boolean)

        Change the current vistrail version into new_version and emit a
        notification signal.

        NB: in most situations, the following post-condition holds:

        >>> controller.change_selected_version(v)
        >>> assert v == controller.current_version

        In some occasions, however, the controller will not be able to
        switch to the desired version. One example where this can
        happen is when the selected version has obsolete modules (that
        is, the currently installed package for those modules has
        module upgrades). In these cases, change_selected_version will
        return a new version which corresponds to a workflow that was
        created by the upgrading mechanism that packages can provide.
        
        """

        try:
            self.do_version_switch(new_version, report_all_errors,
                                   do_validate, from_root)
        except InvalidPipeline, e:
#            from vistrails.gui.application import get_vistrails_application
#
#             def process_err(err):
#                 if isinstance(err, Package.InitializationFailed):
#                     QtGui.QMessageBox.critical(
#                         get_vistrails_application().builderWindow,
#                         'Package load failed',
#                         'Package "%s" failed during initialization. '
#                         'Please contact the developer of that package '
#                         'and report a bug.' % err.package.name)
#                 elif isinstance(err, MissingPackage):
#                     QtGui.QMessageBox.critical(
#                         get_vistrails_application().builderWindow,
#                         'Unavailable package',
#                         'Cannot find package "%s" in\n'
#                         'list of available packages. \n'
#                         'Please install it first.' % err._identifier)
#                 elif issubclass(err.__class__, MissingPort):
#                     msg = ('Cannot find %s port "%s" for module "%s" '
#                            'in loaded package "%s". A different package '
#                            'version might be necessary.') % \
#                            (err._port_type, err._port_name, 
#                             err._module_name, err._package_name)
#                     QtGui.QMessageBox.critical(
#                         get_vistrails_application().builderWindow, 'Missing port',
#                         msg)
#                 else:
#                     QtGui.QMessageBox.critical(
#                         get_vistrails_application().builderWindow,
#                         'Invalid Pipeline', str(err))

            # VisTrails will not raise upgrade exceptions unless
            # configured to do so. To get the upgrade requests,
            # configuration option upgradeModules must be set to True.

            exception_set = e.get_exception_set()
            if len(exception_set) > 0:
#                msg_box = QtGui.QMessageBox(get_vistrails_application().builderWindow)
#                msg_box.setIcon(QtGui.QMessageBox.Warning)
#                msg_box.setText("The current workflow could not be validated.")
#                msg_box.setInformativeText("Errors occurred when trying to "
#                                           "construct this workflow.")
#                msg_box.setStandardButtons(QtGui.QMessageBox.Ok)
#                msg_box.setDefaultButton(QtGui.QMessageBox.Ok)
#                msg_box.setDetailedText(debug.format_exception(e))
#                msg_box.exec_()
                # text = "The current workflow could not be validated."
                # debug.critical(text, e)
                debug.critical("Error changing version", e)

#                 print 'got to exception set'
#                 # Process all errors as usual
#                 if report_all_errors:
#                     for exc in exception_set:
#                         print 'processing', exc
#                         process_err(exc)
#                 else:
#                     process_err(exception_set.__iter__().next())

        except Exception:
            debug.critical('Unexpected Exception',
                           debug.format_exc())
        
        # FIXME: this code breaks undo/redo, and seems to be ok with normal
        # pipeline manipulations so I am leaving it commented out for now

        # if not self._current_terse_graph or \
        #         new_version not in self._current_terse_graph.vertices:
        #     self.recompute_terse_graph()

        self.emit(QtCore.SIGNAL('versionWasChanged'), self.current_version)

    def set_search(self, search, text=''):
        """ set_search(search: SearchStmt, text: str) -> None
        Change the currrent version tree search statement
        
        """
        if self.search != search or self.search_str != text:
            self.search = search
            self.search_str = text
            if self.search:
                self.search.run(self.vistrail, '')
                self.invalidate_version_tree(True)
            if self.refine:
                # need to recompute the graph because the refined items might
                # have changed since last time
                self.recompute_terse_graph()
                self.invalidate_version_tree(True)
            else:
                self.invalidate_version_tree(False)
            
            self.emit(QtCore.SIGNAL('searchChanged'))

    def set_refine(self, refine):
        """ set_refine(refine: bool) -> None
        Set the refine state to True or False
        
        """
        if self.refine!=refine:
            self.refine = refine
            # need to recompute the graph because the refined items might
            # have changed since last time
            self.recompute_terse_graph()
            self.invalidate_version_tree(True)

    def set_full_tree(self, full):
        """ set_full_tree(full: bool) -> None        
        Set if Vistrails should show a complete version tree or just a
        terse tree
        
        """
        if full != self.full_tree:
            self.full_tree = full
            self.invalidate_version_tree(True)

    def recompute_terse_graph(self):
        BaseController.recompute_terse_graph(self)
        self._current_graph_layout.layout_from(self.vistrail,
                                               self._current_terse_graph)

    def refine_graph(self, step=1.0):
        """ refine_graph(step: float in [0,1]) -> (Graph, Graph)        
        Refine the graph of the current vistrail based the search
        status of the controller. It also return the full graph as a
        reference
        
        """
        if self._current_full_graph is None:
            self.recompute_terse_graph()

        return (self._current_terse_graph, self._current_full_graph,
                self._current_graph_layout)

    ##########################################################################
    # undo/redo navigation

    def _change_version_short_hop(self, new_version):
        """_change_version_short_hop is used internally to
        change versions when we're moving exactly one action up or down.
        This allows a few optimizations that improve interactivity."""
        
        if self.current_version <> new_version:
            # Instead of recomputing the terse graph, simply update it

            # There are two variables in play:
            # a) whether or not the destination node is currently on the
            # terse tree (it will certainly be after the move)
            # b) whether or not the current node will be visible (it
            # certainly is now, since it's the current one)

            dest_node_in_terse_tree = new_version in self._current_terse_graph.vertices
            
            current = self.current_version
            tree = self.vistrail.tree.getVersionTree()
            # same logic as recompute_terse_graph except for current
            children_count = len([x for (x, _) in tree.adjacency_list[current]
                                  if (x in self.vistrail.actionMap and
                                      not self.vistrail.is_pruned(x))])
            current_node_will_be_visible = \
                (self.full_tree or
                 self.vistrail.has_tag(self.current_version) or
                 children_count <> 1)

            self.change_selected_version(new_version)
            # case 1:
            if not dest_node_in_terse_tree and \
                    not current_node_will_be_visible and not current == 0:
                # we're going from one boring node to another,
                # so just rename the node on the terse graph
                self._current_terse_graph.rename_vertex(current, new_version)
                self.replace_unnamed_node_in_version_tree(current, new_version)
            else:
                # bail, for now
                self.recompute_terse_graph()
                self.invalidate_version_tree(False)
        

    def show_parent_version(self):
        """ show_parent_version() -> None
        Go back one from the current version and display it

        """
        # NOTE cscheid: Slight change in the logic under refined views:
        # before r1185, undo would back up more than one action in the
        # presence of non-matching refined nodes. That seems wrong. Undo
        # should always move one step only.         

        prev = None
        try:
            prev = self._current_full_graph.parent(self.current_version)
        except Graph.VertexHasNoParentError:
            prev = 0

        self._change_version_short_hop(prev)

    def show_child_version(self, which_child):
        """ show_child_version(which_child: int) -> None
        Go forward one version and display it. This is used in redo.

        ONLY CALL THIS FUNCTION IF which_child IS A CHILD OF self.current_version

        """
        self._change_version_short_hop(which_child)

    def setSavedQueries(self, queries):
        """ setSavedQueries(queries: list of (str, str, str)) -> None
        Set the saved queries of a vistail
        
        """
        self.vistrail.setSavedQueries(queries)
        self.set_changed(True)
        
    def update_current_tag(self, tag):
        """ update_current_tag(tag: str) -> Bool
        Update the current vistrail tag and return success predicate
        
        """
        self.flush_delayed_actions()

        # Find current location of tag
        tag_version = self.vistrail.search_upgrade_versions(
                self.current_version,
                lambda vt, v, bv: v if vt.has_tag(v) else None)

        # No change in tag: return
        if (tag_version is not None and
                self.vistrail.getVersionName(tag_version) == tag):
            return True

        # This tag already exists elsewhere: error
        if self.vistrail.has_tag_str(tag):
            show_warning("Name Exists",
                         "There is another version named '%s'.\n"
                         "Please enter a different name." % tag)
            return False

        # Remove current tag
        self.vistrail.set_tag(tag_version, None)

        if self.vistrail.hasTag(self.current_version):
            self.vistrail.changeTag(tag, self.current_base_version)
        else:
            self.vistrail.addTag(tag, self.current_base_version)

        self.set_changed(True)
        self.recompute_terse_graph()
        self.invalidate_version_tree(False)
        return True

    def perform_param_changes(self, actions):
        """perform_param_changes(actions) -> None

        Performs a series of parameter change actions to the current version.

        FIXME: this function seems to be called from a single place in
        the spreadsheet cell code. Do we need it?
        """
        if len(actions) == 0:
            return
        for action in actions:
            for operation in action.operations:
                if operation.vtType == 'add' or operation.vtType == 'change':
                    if operation.new_obj_id < 0:
                        data = operation.data
                        new_id = self.vistrail.idScope.getNewId(data.vtType)
                        data.real_id = new_id
                        operation.new_obj_id = new_id
            self.add_new_action(action)
            self.perform_action(action, quiet=True)
        self.set_changed(True)
        self.invalidate_version_tree(False)

    def get_pipeline_name(self, version=None):
        if version is None:
            version = self.current_version
        return self.vistrail.get_pipeline_name(version)

    ###########################################################################
    # Clipboard, copy/paste

    def get_selected_item_ids(self):
        return self.current_pipeline_scene.get_selected_item_ids()

    def copy_modules_and_connections(self, module_ids, connection_ids):
        """copy_modules_and_connections(module_ids: [long],
                                     connection_ids: [long]) -> str
        Serializes a list of modules and connections
        """
        self.flush_delayed_actions()

        def process_group(group):
            # reset pipeline id for db
            group.pipeline.id = None
            # recurse
            for module in group.pipeline.module_list:
                if module.is_group():
                    process_group(module)

        pipeline = Pipeline()
        sum_x = 0.0
        sum_y = 0.0
        for module_id in module_ids:
            module = self.current_pipeline.modules[module_id]
            sum_x += module.location.x
            sum_y += module.location.y
            if module.is_group():
                process_group(module)

        center_x = sum_x / len(module_ids)
        center_y = sum_y / len(module_ids)
        for module_id in module_ids:
            module = self.current_pipeline.modules[module_id]
            module = module.do_copy()
            module.location.x -= center_x
            module.location.y -= center_y
            pipeline.add_module(module)
        for connection_id in connection_ids:
            connection = self.current_pipeline.connections[connection_id]
            pipeline.add_connection(connection)
        return vistrails.core.db.io.serialize(pipeline)
        
    def paste_modules_and_connections(self, str, center):
        """ paste_modules_and_connections(str,
                                          center: (float, float)) -> [id list]
        Paste a list of modules and connections into the current pipeline.

        Returns the list of module ids of added modules

        """
        def remove_duplicate_aliases(pip):
            aliases = self.current_pipeline.aliases.keys()
            for a in aliases:
                if a in pip.aliases:
                    (type, oId, parentType, parentId, mid) = pip.aliases[a]
                    pip.remove_alias_by_name(a)
                    _mod = pip.modules[mid]
                    _fun = _mod.function_idx[parentId]
                    _par = _fun.parameter_idx[oId]
                    _par.alias = ''
                                    
        self.flush_delayed_actions()
        pipeline = vistrails.core.db.io.unserialize(str, Pipeline)
        remove_duplicate_aliases(pipeline)

        modules = []
        if pipeline:
            def process_group(group):
                # reset pipeline id for db
                group.pipeline.id = None
                # recurse
                for module in group.pipeline.module_list:
                    if module.is_group():
                        process_group(module)

            for module in pipeline.module_list:
                module.location.x += center[0]
                module.location.y += center[1]
                if module.is_group():
                    process_group(module)

            id_remap = {}
            action = vistrails.core.db.action.create_paste_action(pipeline, 
                                                        self.vistrail.idScope,
                                                        id_remap)

            modules = [op.objectId
                       for op in action.operations
                       if (op.what == 'module' or 
                           op.what == 'abstraction' or
                           op.what == 'group')]

            self.add_new_action(action)
            self.vistrail.change_description("Paste", action.id)
            self.perform_action(action)
            self.validate(self.current_pipeline, False)
        return modules

    def get_abstraction_name(self, name="", check_exists=True):
        name = self.do_abstraction_prompt(name)
        if name is None:
            return None
        while name == "" or (check_exists and self.abstraction_exists(name)):
            name = self.do_abstraction_prompt(name, name != "")
            if name is None:
                return None
        return name

    def do_abstraction_prompt(self, name="", exists=False):
        if exists:
            prompt = "'%s' already exists.  Enter a new subworkflow name" % \
                name
        else:
            prompt = 'Enter subworkflow name'
            
        (text, ok) = QtGui.QInputDialog.getText(None, 
                                                'Set SubWorkflow Name',
                                                prompt,
                                                QtGui.QLineEdit.Normal,
                                                name)
        if ok and text:
            return str(text).strip().rstrip()
        if not ok:
            return None
        return ""

    def import_abstractions(self, abstraction_ids):
        for abstraction_id in abstraction_ids:
            abstraction = self.current_pipeline.modules[abstraction_id]
            new_name = self.get_abstraction_name(abstraction.name)
            if new_name:
                self.import_abstraction(new_name,
                                        abstraction.package,
                                        abstraction.name, 
                                        abstraction.namespace,
                                        abstraction.internal_version)
        
    def do_export_prompt(self, title, prompt):
        (text, ok) = QtGui.QInputDialog.getText(None,
                                                title,
                                                prompt,
                                                QtGui.QLineEdit.Normal,
                                                '')
        if ok and not text:
            return str(text).strip().rstrip()
        return ''
            
    def do_save_dir_prompt(self):
        dialog = QtGui.QFileDialog.getExistingDirectory
        dir_name = dialog(None, "Save Subworkflows...",
                          vistrails.core.system.vistrails_file_directory())
        if dir_name:
            return None
        dir_name = os.path.abspath(str(dir_name))
        setattr(get_vistrails_configuration(), 'fileDir', dir_name)
        vistrails.core.system.set_vistrails_file_directory(dir_name)
        return dir_name

    def create_abstractions_from_groups(self, group_ids):
        for group_id in group_ids:
            self.create_abstraction_from_group(group_id)

    def create_abstraction_from_group(self, group_id, name=""):
        self.flush_delayed_actions()
        name = self.get_abstraction_name(name)

        (abstraction, connections) = \
            self.build_abstraction_from_group(self.current_pipeline,
                                              group_id, name)

        op_list = []
        getter = self.get_connections_to_and_from
        op_list.extend(('delete', c)
                       for c in getter(self.current_pipeline, [group_id]))
        op_list.append(('delete', self.current_pipeline.modules[group_id]))
        op_list.append(('add', abstraction))
        op_list.extend(('add', c) for c in connections)
        action = vistrails.core.db.action.create_action(op_list)
        self.add_new_action(action)
        result = self.perform_action(action)
        return abstraction

    def export_abstractions(self, abstraction_ids):
        save_dir = self.do_save_dir_prompt()
        if not save_dir:
            return 

        def read_init(dir_name):
            import imp
            found_attrs = {}
            found_lists = {}
            attrs = ['identifier', 'name', 'version']
            lists = ['_subworkflows', '_dependencies']
            try:
                (file, pathname, description) = \
                    imp.find_module(os.path.basename(dir_name), 
                                    [os.path.dirname(dir_name)])
                module = imp.load_module(os.path.basename(dir_name), file,
                                         pathname, description)
                for attr in attrs:
                    if hasattr(module, attr):
                        found_attrs[attr] = getattr(module, attr)
                for attr in lists:
                    if hasattr(module, attr):
                        found_lists[attr] = getattr(module, attr)
            except Exception, e:
                debug.critical("Exception: %s" % e)
                pass
            return (found_attrs, found_lists)

        def write_init(save_dir, found_attrs, found_lists, attrs, lists):
            init_file = os.path.join(save_dir, '__init__.py')
            if os.path.exists(init_file):
                f = open(init_file, 'a')
            else:
                f = open(init_file, 'w')
            for attr, val in attrs.iteritems():
                if attr not in found_attrs:
                    print >>f, "%s = '%s'" % (attr, val)
            for attr, val_list in lists.iteritems():
                if attr not in found_lists:
                    print >>f, "%s = %s" % (attr, str(val_list))
                else:
                    diff_list = []
                    for val in val_list:
                        if val not in found_lists[attr]:
                            diff_list.append(val)
                    print >>f, '%s.extend(%s)' % (attr, str(diff_list))
            f.close()

        if os.path.exists(os.path.join(save_dir, '__init__.py')):
            (found_attrs, found_lists) = read_init(save_dir)
        else:
            found_attrs = {}
            found_lists = {}

        if 'name' in found_attrs:
            pkg_name = found_attrs['name']
        else:
            pkg_name = self.do_export_prompt("Target Package Name",
                                             "Enter target package name")
            if not pkg_name:
                return

        if 'identifier' in found_attrs:
            pkg_identifier = found_attrs['identifier']
        else:
            pkg_identifier = self.do_export_prompt("Target Package Identifier",
                                                   "Enter target package "
                                                   "identifier (e.g. "
                                                   "org.place.user.package)")
            if not pkg_identifier:
                return

        abstractions = []
        for abstraction_id in abstraction_ids:
            abstraction = self.current_pipeline.modules[abstraction_id]
            if abstraction.is_abstraction() and \
                    abstraction.package == abstraction_pkg:
                abstractions.append(abstraction)
                for v in self.find_abstractions(abstraction.vistrail).itervalues():
                    abstractions.extend(v)
        pkg_subworkflows = []
        pkg_dependencies = set()
        for abstraction in abstractions:
            new_name = self.get_abstraction_name(abstraction.name, False)
            if not new_name:
                break
            (subworkflow, dependencies) = \
                self.export_abstraction(new_name,
                                        pkg_identifier,
                                        save_dir,
                                        abstraction.package,
                                        abstraction.name, 
                                        abstraction.namespace,
                                        str(abstraction.internal_version))
            pkg_subworkflows.append(subworkflow)
            pkg_dependencies.update(dependencies)

        attrs = {'identifier': pkg_identifier,
                 'name': pkg_name,
                 'version': '0.0.1'}
        lists = {'_subworkflows': pkg_subworkflows,
                 '_dependencies': list(pkg_dependencies)}
        write_init(save_dir, found_attrs, found_lists, attrs, lists)

    def set_changed(self, changed):
        """ set_changed(changed: bool) -> None
        Set the current state of changed and emit signal accordingly
        
        """
        BaseController.set_changed(self, changed)
        if changed:
            # FIXME: emit different signal in the future
            self.emit(QtCore.SIGNAL('stateChanged'))

    def set_file_name(self, file_name):
        """ set_file_name(file_name: str) -> None
        Change the controller file name
        
        """
        old_name = self.file_name
        BaseController.set_file_name(self, file_name)
        if old_name!=file_name:
            self.emit(QtCore.SIGNAL('stateChanged'))

    def write_vistrail(self, locator, version=None, export=False):
        need_invalidate = BaseController.write_vistrail(self, locator,
                                                        version, export)
        if need_invalidate and not export:
            self.invalidate_version_tree(False)
            self.set_changed(False)

    def write_opm(self, locator):
        if self.log:
            if self.vistrail.db_log_filename is not None:
                log = vistrails.core.db.io.merge_logs(self.log, 
                                            self.vistrail.db_log_filename)
            else:
                log = self.log
            opm_graph = OpmGraph(log=log, 
                                 version=self.current_version,
                                 workflow=self.current_pipeline,
                                 registry=get_module_registry())
            locator.save_as(opm_graph)
            
    def write_prov(self, locator):
        if self.log:
            if self.vistrail.db_log_filename is not None:
                log = vistrails.core.db.io.merge_logs(self.log, 
                                            self.vistrail.db_log_filename)
            else:
                log = self.log
            prov_document = ProvDocument(log=log, 
                                         version=self.current_version,
                                         workflow=self.current_pipeline,
                                         registry=get_module_registry())
            locator.save_as(prov_document)

    def query_by_example(self, pipeline):
        """ query_by_example(pipeline: Pipeline) -> None
        Perform visual query on the current vistrail
        
        """
        if len(pipeline.modules)==0:
            search = TrueSearch()
        else:
            if not self._current_terse_graph:
                self.recompute_terse_graph()
            versions_to_check = \
                set(self._current_terse_graph.vertices.iterkeys())
            search = VisualQuery(pipeline, versions_to_check)

        self.set_search(search, '') # pipeline.dump_to_string())

    ##########################################################################
    # analogies

    def add_analogy(self, analogy_name, version_from, version_to):
        assert isinstance(analogy_name, str)
        assert isinstance(version_from, (int, long))
        assert isinstance(version_to, (int, long))
        if analogy_name in self.analogy:
            raise VistrailsInternalError("duplicated analogy name '%s'" %
                                         analogy_name)
        self.analogy[analogy_name] = (version_from, version_to)

    def remove_analogy(self, analogy_name):
        if analogy_name not in self.analogy:
            raise VistrailsInternalError("missing analogy '%s'" %
                                         analogy_name)
        del self.analogy[analogy_name]

    def perform_analogy(self, analogy_name, analogy_target):
        if analogy_name not in self.analogy:
            raise VistrailsInternalError("missing analogy '%s'" %
                                         analogy_name)

        # remove delayed actions since we're not necessarily using
        # current_version
        self._delayed_actions = []

        (a, b) = self.analogy[analogy_name]
        c = analogy_target
        if self.current_version != c:
            self.change_selected_version(c)

        try:
            pipeline_a = self.vistrail.getPipeline(a)
            self.validate(pipeline_a)
        except InvalidPipeline, e:
            (_, pipeline_a) = \
                self.handle_invalid_pipeline(e, a, Vistrail())
            self._delayed_actions = []
        try:
            pipeline_c = self.vistrail.getPipeline(c)
            self.validate(pipeline_c)
        except InvalidPipeline, e:
            (_, pipeline_c) = self.handle_invalid_pipeline(e, a, Vistrail())
            self._delayed_actions = []
                                                     
        action = vistrails.core.analogy.perform_analogy_on_vistrail(self.vistrail,
                                                          a, b, c, 
                                                          pipeline_a,
                                                          pipeline_c)
        self.add_new_action(action)
        self.vistrail.change_description("Analogy", action.id)
        self.vistrail.change_analogy_info("(%s -> %s)(%s)" % (a, b, c), 
                                          action.id)
        
        # make sure that the output from the analogy is as up-to-date
        # as we can make it
        self.change_selected_version(action.id, from_root=True)
        self.flush_delayed_actions()
        self.invalidate_version_tree()
        
    def executeParameterExploration(self, pe, view=None, extra_info={}, showProgress=True):
        """ execute(pe: ParameterExploration, view: QVistrailView,
            extra_info: dict, showProgress: bool) -> None
        Perform the exploration by collecting a list of actions
        corresponding to each dimension
        
        """
        reg = get_module_registry()
        spreadsheet_pkg = 'org.vistrails.vistrails.spreadsheet'
        use_spreadsheet = reg.has_module(spreadsheet_pkg, 'CellLocation') and\
                          reg.has_module(spreadsheet_pkg, 'SheetReference')

        if pe.action_id != self.current_version:
            self.change_selected_version(pe.action_id)
        actions, pre_actions, vistrail_vars = \
                        pe.collectParameterActions(self.current_pipeline)

        if self.current_pipeline and actions:
            pe_log_id = uuid.uuid1()
            explorer = ActionBasedParameterExploration()
            (pipelines, performedActions) = explorer.explore(
                self.current_pipeline, actions, pre_actions)
            
            dim = [max(1, len(a)) for a in actions]
            if use_spreadsheet:
                from vistrails.gui.paramexplore.virtual_cell import positionPipelines, assembleThumbnails
                from vistrails.gui.paramexplore.pe_view import QParamExploreView
                modifiedPipelines, pipelinePositions = positionPipelines(
                    'PE#%d %s' % (QParamExploreView.explorationId, self.name),
                    dim[2], dim[1], dim[0], pipelines, pe.layout, self)
                QParamExploreView.explorationId += 1
            else:
                from vistrails.core.param_explore import _pipelinePositions
                modifiedPipelines = pipelines
                pipelinePositions = _pipelinePositions(
                    dim[2], dim[1], dim[0], pipelines)

            mCount = []
            for p in modifiedPipelines:
                if len(mCount)==0:
                    mCount.append(0)
                else:
                    mCount.append(len(p.modules)+mCount[len(mCount)-1])
                
            # Now execute the pipelines
            if showProgress:
                totalProgress = sum([len(p.modules) for p in modifiedPipelines])
                progress = QtGui.QProgressDialog('Performing Parameter '
                                                 'Exploration...',
                                                 '&Cancel',
                                                 0, totalProgress)
                progress.setWindowTitle('Parameter Exploration')
                progress.setWindowModality(QtCore.Qt.WindowModal)
                progress.show()

            interpreter = get_default_interpreter()
            
            images = {}
            errors = []
            for pi in xrange(len(modifiedPipelines)):
                if showProgress:
                    progress.setValue(mCount[pi])
                    QtCore.QCoreApplication.processEvents()
                    if progress.wasCanceled():
                        break
                    def moduleExecuted(objId):
                        if not progress.wasCanceled():
                            progress.setValue(progress.value()+1)
                            QtCore.QCoreApplication.processEvents()
                if use_spreadsheet:
                    name = os.path.splitext(self.name)[0] + \
                                         ("_%s_%s_%s" % pipelinePositions[pi])
                    extra_info['nameDumpCells'] = name
                    if 'pathDumpCells' in extra_info:
                        images[pipelinePositions[pi]] = \
                                   os.path.join(extra_info['pathDumpCells'], name)
                pe_cell_id = (pe_log_id,) + pipelinePositions[pi]
                kwargs = {'locator': self.locator,
                          'current_version': self.current_version,
                          'reason': 'Parameter Exploration %s %s_%s_%s' % pe_cell_id,
                          'logger': self.get_logger(),
                          'actions': performedActions[pi],
                          'extra_info': extra_info
                          }
                if view:
                    kwargs['view'] = view
                if showProgress:
                    kwargs['module_executed_hook'] = [moduleExecuted]
                if self.get_vistrail_variables():
                    # remove vars used in pe
                    vars = dict([(v.uuid, v) for v in self.get_vistrail_variables()
                            if v.uuid not in vistrail_vars])
                    kwargs['vistrail_variables'] = lambda x: vars.get(x, None)
                result = interpreter.execute(modifiedPipelines[pi], **kwargs)
                for error in result.errors.itervalues():
                    if use_spreadsheet:
                        pp = pipelinePositions[pi]
                        errors.append(((pp[1], pp[0], pp[2]), error))
                    else:
                        errors.append(((0,0,0), error))

            if showProgress:
                progress.setValue(totalProgress)
            if 'pathDumpCells' in extra_info:
                filename = os.path.join(extra_info['pathDumpCells'],
                                        os.path.splitext(self.name)[0])
                assembleThumbnails(images, filename)
            from vistrails.gui.vistrails_window import _app
            _app.notify('execution_updated')
            return errors

################################################################################
# Testing


class TestVistrailController(vistrails.gui.utils.TestVisTrailsGUI):

    # def test_add_module(self):
    #     v = api.new_vistrail()
       
    def tearDown(self):
        vistrails.gui.utils.TestVisTrailsGUI.tearDown(self)

        d = vistrails.core.system.get_vistrails_directory('subworkflowsDir')
        filename = os.path.join(d, '__TestFloatList.xml')
        if os.path.exists(filename):
            os.remove(filename)

    def test_create_functions(self):
        controller = VistrailController(Vistrail(), None, 
                                        pipeline_view=DummyView(),
                                        auto_save=False)
        controller.change_selected_version(0L)
        module = controller.add_module(0.0,0.0, 
                        vistrails.core.system.get_vistrails_basic_pkg_id(), 
                        'ConcatenateString')
        functions = [('str1', ['foo'], -1, True),
                     ('str2', ['bar'], -1, True)]
        controller.update_functions(module, functions)

        self.assertEquals(len(controller.current_pipeline.module_list), 1)
        p_module = controller.current_pipeline.modules[module.id]
        self.assertEquals(len(p_module.functions), 2)
        self.assertEquals(p_module.functions[0].params[0].strValue, 'foo')
        self.assertEquals(p_module.functions[1].params[0].strValue, 'bar')

        # make sure updates work correctly
        # also check that we can add more than one function w/ same name
        # by passing False as should_replace
        new_functions = [('str1', ['baz'], -1, True),
                         ('str2', ['foo'], -1, False),
                         ('str3', ['bar'], -1, False)]
        controller.update_functions(p_module, new_functions)
        self.assertEquals(len(p_module.functions), 4)

    def test_abstraction_create(self):
        from vistrails.core.db.locator import XMLFileLocator
        d = vistrails.core.system.get_vistrails_directory('subworkflowsDir')
        filename = os.path.join(d, '__TestFloatList.xml')
        locator = XMLFileLocator(vistrails.core.system.vistrails_root_directory() +
                           '/tests/resources/test_abstraction.xml')
        v = locator.load()
        controller = VistrailController(v, locator, pipeline_view=DummyView(),
                                        auto_save=False)
        # DAK: version is different because of upgrades
        # controller.change_selected_version(9L)
        controller.select_latest_version()
        self.assertNotEqual(controller.current_pipeline, None)

        # If getting a KeyError here, run the upgrade on the vistrail and
        # update the ids
        # TODO : rewrite test so we don't have to update this unrelated code
        # each time new upgrades are introduced
        # Original ids:
        #     module_ids = [1, 2, 3]
        #     connection_ids = [1, 2, 3]
        module_ids = [15, 13, 14]
        connection_ids = [21, 18, 20]
        controller.create_abstraction(module_ids, connection_ids,
                                      '__TestFloatList')
        self.assert_(os.path.exists(filename))



###############################################################################
##
## Copyright (C) 2014-2015, New York University.
## Copyright (C) 2011-2014, NYU-Poly.
## Copyright (C) 2006-2011, University of Utah.
## All rights reserved.
## Contact: contact@vistrails.org
##
## This file is part of VisTrails.
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are met:
##
##  - Redistributions of source code must retain the above copyright notice,
##    this list of conditions and the following disclaimer.
##  - Redistributions in binary form must reproduce the above copyright
##    notice, this list of conditions and the following disclaimer in the
##    documentation and/or other materials provided with the distribution.
##  - Neither the name of the New York University nor the names of its
##    contributors may be used to endorse or promote products derived from
##    this software without specific prior written permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
## AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
## THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
## PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
## CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
## EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
## PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
## OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
## WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
## OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
## ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
##
###############################################################################
""" This file contains classes related to loading and saving a set of
connection objects to an XML file.
It defines the following
classes:
 - ExtConnection
 - DBConnection
 - ExtConnectionList
"""
from __future__ import division

import os
import tempfile
import unittest

from vistrails.core.utils import VistrailsInternalError, abstract
from vistrails.core.utils.enum import enum
from vistrails.core.utils.uxml import named_elements, XMLWrapper
################################################################################

ConnectionType = enum('ConnectionType',
                      ['DB', 'HTTP', 'Unknown'],
                      "Enumeration of Connection Types")

class ExtConnection(object):
    """Stores Information for an External Connection"""
    parseDispatch = {}
    
    def __init__(self, id=-1, name='', host='', type=ConnectionType.Unknown):
        """__init__(id: int,  name: str, host: str, type: ConnectionType)
                    -> ExtConnection
        It creates an external connection. Ignoring the connection type, every
        connection should have at least a name, a hostname and an id """
        self.id = id
        self.name = name
        self.host = host
        self.type = type
        
    def serialize(self, dom, element):
        abstract()

    @staticmethod
    def parse(element):
        type = str(element.getAttribute('type'))
        return ExtConnection.parseDispatch[type](element)

class DBConnection(ExtConnection):
    """Stores Information for Database Connection """
    def __init__(self, id=-1, name='', host='', port=0, user='', passwd='',
                 database='', dbtype=''):
        """__init__(id: int,  name: str, host: str, port: int, username: str,
                    passwd: str, database: str, type:ConnectionType)->DBConnection
           It creates a DBConnection Object
        """
        ExtConnection.__init__(self, id, name, host, ConnectionType.DB)
        self.port = port
        self.user = user
        self.passwd = passwd
        self.database = database
        self.dbtype = dbtype

    def serialize(self, dom, element):
        """serialize(dom, element) -> None
        Convert this object to an XML representation.
        """
        conn = dom.createElement('connection')
        conn.setAttribute('id', str(self.id))
        conn.setAttribute('name', str(self.name))
        conn.setAttribute('host', str(self.host))
        conn.setAttribute('port',str(self.port))
        conn.setAttribute('user', str(self.user))
        conn.setAttribute('passwd', str(self.passwd))
        conn.setAttribute('database', str(self.database))
        conn.setAttribute('type', str(self.type))
        conn.setAttribute('dbtype', str(self.dbtype))
        element.appendChild(conn)

    def __str__(self):
        """ __str__() -> str - Writes itself as a string """ 
        return """<<id= '%s' name='%s' type='%s' host='%s' 
        user='%s' database='%s' dbtype='%s'>>""" %  (
            self.id,
            self.name,
            self.type,
            self.host,
            self.user,
            self.database,
            self.dbtype)

    @staticmethod
    def parse(element):
        """ parse(element) -> DBConnection
        Parse an XML object representing a DBConnection and returns a
        DBConnection object. 
        
        """
        conn = DBConnection()
        conn.type = ConnectionType.DB
        conn.id = int(element.getAttribute('id'))
        conn.name = str(element.getAttribute('name'))
        conn.host = str(element.getAttribute('host'))
        conn.port = int(element.getAttribute('port'))
        conn.user = str(element.getAttribute('user'))
        conn.passwd = str(element.getAttribute('passwd'))
        conn.database = str(element.getAttribute('database'))
        conn.dbtype = str(element.getAttribute('dbtype'))
        return conn

    def __eq__(self, other):
        """ __eq__(other: DBConnection) -> boolean
        Returns True if self and other have the same attributes. Used by == 
        operator. 
        
        """
        if other is None:
            return False
        if self.type != other.type:
            return False
        if self.id != other.id:
            return False
        if self.name != other.name:
            return False
        if self.host != other.host:
            return False
        if self.port != other.port:
            return False
        if self.user != other.user:
            return False
        if self.passwd != other.passwd:
            return False
        if self.database != other.database:
            return False
        if self.dbtype != other.dbtype:
            return False
        return True

    def __ne__(self, other):
        return not (self == other)

ExtConnection.parseDispatch[str(ConnectionType.DB)] = DBConnection.parse

class ExtConnectionList(XMLWrapper):
    """Class to store and manage a list of connections.

    """
    _instance = None
    @staticmethod
    def getInstance(*args, **kwargs):
        if ExtConnectionList._instance is None:
            obj = ExtConnectionList(*args, **kwargs)
            ExtConnectionList._instance = obj
        return ExtConnectionList._instance
    
    def __init__(self, filename=''):
        """ __init__() -> ExtConnectionList """
        if not ExtConnectionList._instance:
            self.__connections = {}
            self.changed = False
            self.current_id = 1
            self.filename = filename
            self.load_connections()
            ExtConnectionList._instance = self
        else:
            raise RuntimeError, 'Only one instance of ExtConnectionList is \
allowed!'

    def load_connections(self):
        """load_connections()-> None
        Load connections from its internal filename

        """
        if os.path.exists(self.filename):
            self.parse(self.filename)

    def add_connection(self, conn):
        """add_connection(conn: ExtConnection) -> None
        Adds a connection to the list

        """
        if self.__connections.has_key(conn.id):
            msg = "External Connection '%s' with repeated id" % conn.name
            raise VistrailsInternalError(msg)
        self.__connections[conn.id] = conn
        self.current_id = max(self.current_id, conn.id+1)
        self.serialize()

    def get_connection(self, id):
        """get_connection(id: int) -> ExtConnection
        Returns connection object associated with id

        """
        if self.__connections.has_key(id):
            return self.__connections[id]
        else:
            return None

    def has_connection(self, id):
        """has_connection(id: int) -> Boolean
        Returns True if connection with id exists """
        return self.__connections.has_key(id)

    def find_db_connection(self, host, port, db):
        """find_db_connection(host: str, port: int, db: str) -> id
        Returns the id of the first connection that matches the provided
        parameters. It will return -1 if not found

        """
        for conn in self.__connections.itervalues():
            if conn.host == host and conn.port == port and conn.database == db:
                return conn.id
        return -1
    
    def set_connection(self, id, conn):
        """set_connection(id: int, conn: ExtConnection)- > None
        Updates the connection with id to be conn

        """
        if self.__connections.has_key(id):
            self.__connections[id] = conn
            self.serialize()
            
    def remove_connection(self, id):
        """remove_connection(id: int) -> None 
        Remove connection with id 'id'
        
        """
        if self.__connections.has_key(id):
            del self.__connections[id]
            self.serialize()
        
    def clear(self):
        """ clear() -> None 
        Remove current connections """
        self.__connections.clear()
        self.current_id = 1

    def count(self):
        """count() -> int - Returns the number of connections """
        return len(self.__connections)

    def items(self):
        """ items() -> - Returns the connections """
        return self.__connections.items()

    def parse(self, filename):
        """parse(filename: str) -> None  
        Loads a list of connections from a XML file, appending it to
        self.__connections.
        
        """
        self.open_file(filename)
        root = self.dom.documentElement
        for element in named_elements(root, 'connection'):
            self.add_connection(ExtConnection.parse(element))
        self.refresh_current_id()

    def serialize(self):
        """serialize(filename:str) -> None 
        Writes connection list to given filename.
          
        """
        dom = self.create_document('connections')
        root = dom.documentElement
        
        for conn in self.__connections.values():
            conn.serialize(dom, root)

        self.write_document(root, self.filename)

    def refresh_current_id(self):
        """refresh_current_id() -> None
        Recomputes the next unused id from scratch
        
        """
        self.current_id = max([0] + self.__connections.keys()) + 1

    def get_fresh_id(self):
        """get_fresh_id() -> int - Returns an unused id. """
        return self.current_id


###############################################################################

class TestConnectionList(unittest.TestCase):
    def test1(self):
        """ Exercising writing and reading a file """
        fd, filename = tempfile.mkstemp(prefix='vt_', suffix='_connections.xml')
        os.close(fd)
        try:
            conns = ExtConnectionList.getInstance()
            conns.filename = filename
            conns.clear()
            conn = DBConnection()
            conn.id = 1
            conn.name = 'test'
            conn.host = 'somehost.com'
            conn.port = 1234
            conn.user = 'nobody'
            conn.passwd = '123'
            conn.database = 'anydatabase'
            conn.dbtype = 'MySQL'

            conns.add_connection(conn)

            #reading it again
            conns.clear()
            self.assertEquals(conns.count(),0)
            conns.load_connections()
            self.assertEquals(conns.count(),1)
            newconn = conns.get_connection(1)
            self.assertEqual(conn, newconn)
        finally:
            #remove created file
            os.unlink(filename)

if __name__ == '__main__':
    unittest.main()
